#! /usr/bin/env python
"""Utilities to aid interaction with the unicode database"""

import logging
import os.path

from sys import maxunicode
from pickle import dump, load

from .py2 import (
    byte,
    byte_value,
    character,
    dict_values,
    force_text,
    is_text,
    is_unicode,
    join_bytes,
    py2,
    range3,
    suffix,
    to_text,
    ul,
    UnicodeMixin,
    urlopen)
from .pep8 import PEP8Compatibility


UCDDatabaseURL = "http://www.unicode.org/Public/UNIDATA/UnicodeData.txt"
UCDBlockDatabaseURL = "http://www.unicode.org/Public/UNIDATA/Blocks.txt"
UCDCategories = {}
UCDBlocks = {}

CATEGORY_FILE = "unicode5_catogories%s.pck" % suffix
BLOCK_FILE = "unicode5_blocks%s.pck" % suffix


MAGIC_TABLE = {
    # UCS-4, big-endian machine (1234 order)
    b'\x00\x00\xfe\xff': 'utf_32_be',
    # UCS-4, little-endian (4321 order)
    b'\xff\xfe\x00\x00': 'utf_32_le',
    # UCS-4, unusual octet order (2143)
    b'\x00\x00\xff\xfe': 'utf_32',
    # UCS-4, unusual octet order (3412)
    b'\xfe\xff\x00\x00': 'utf_32',
    # UTF-16, big-endian
    b'\xfe\xff\x00\x2a': 'utf_16_be',
    # UTF-16, little-endian
    b'\xff\xfe\x2a\x00': 'utf_16_le',
    # UCS-4 or other encoding with a big-endian 32-bit code unit
    b'\x00\x00\x00\x2a': 'utf_32_be',
    # UCS-4 or other encoding with a little-endian 32-bit code unit
    b'\x2a\x00\x00\x00': 'utf_32_le',
    # UCS-4 or other encoding with an unusual 32-bit code unit
    b'\x00\x00\x2a\x00': 'utf_32_le',
    # UCS-4 or other encoding with an unusual 32-bit code unit
    b'\x00\x2a\x00\x00': 'utf_32_le',
    # UTF-16BE or big-endian ISO-10646-UCS-2 or other encoding with a
    # 16-bit code unit
    b'\x00\x2a\x00\x2a': 'utf_16_be',
    # UTF-16LE or little-endian ISO-10646-UCS-2 or other encoding with a
    # 16-bit code unit
    b'\x2a\x00\x2a\x00': 'utf_16_le',
    # UTF-8, ISO 646, ASCII or similar
    b'\x2a\x2a\x2a\x2a': 'utf_8'
}


def detect_encoding(magic):
    """Detects text encoding

    magic
        A string of bytes

    Given a byte string this function looks at (up to) four bytes and
    returns a best guess at the unicode encoding being used for the
    data.

    It returns a string suitable for passing to Python's native decode
    method, e.g., 'utf_8'.  The default is 'utf_8', an encoding which
    will also work if the data is plain ASCII."""
    if magic[0:3] == b'\xef\xbb\xbf':
        # catch this odd one first
        return 'utf_8_sig'
    # now we're only interested in a handful of values
    keepers = b"\x00\xfe\xff"
    key = []
    for i in range3(4):
        # star - as good as any
        if i >= len(magic):
            key.append(byte(0x2A))
        elif magic[i] in keepers:
            key.append(magic[i])
        else:
            key.append(byte(0x2A))
    key = join_bytes(key)
    if key in MAGIC_TABLE:
        return MAGIC_TABLE[key]
    else:
        return None


class CharClass(UnicodeMixin):

    """Represents a class of unicode characters.

    A class of characters is represented internally by a list of
    character ranges that define the class.  This is efficient because
    most character classes are defined in blocks of characters.

    For the constructor, multiple arguments can be provided.

    String arguments add all characters in the string to the class.  For
    example, CharClass('abcxyz') creates a class comprising two ranges:
    a-c and x-z.

    Tuple/List arguments can be used to pass pairs of characters that
    define a range.  For example, CharClass(('a','z')) creates a class
    comprising the letters a-z.

    Instances of CharClass can also be used in the constructor to add an
    existing class.

    Instances support Python's repr function::

        >>> c = CharClass('abcxyz')
        >>> print repr(c)
        CharClass((u'a',u'c'), (u'x',u'z'))

    The string representation of a CharClass is a python regular
    expression suitable for matching a single character from the
    CharClass::

        >>> print str(c)
        [a-cx-z]

    """

    @classmethod
    def ucd_category(cls, category):
        """Returns the character class representing the Unicode category.

        You must *not* modify the returned instance, if you want to
        derive a character class from one of the standard Unicode
        categories then you should create a copy by passing the result
        of this class method to the CharClass constructor, e.g. to
        create a class of all general controls and the space character::

                c=CharClass(CharClass.ucd_category(u"Cc"))
                c.add_char(u" ")"""
        global UCDCategories
        if not UCDCategories:
            # The category table is empty, so we need to load it
            load_category_table()
        return UCDCategories[category]

    @classmethod
    def ucd_block(cls, block_name):
        """Returns the character class representing the Unicode block.

        You must not modify the returned instance, if you want to derive
        a character class from one of the standard Unicode blocks then
        you should create a copy by passing the result of this class
        method to the CharClass constructor, e.g. to create a class
        combining all Basic Latin characters and those in the Latin-1
        Supplement::

            c=CharClass(CharClass.ucd_block(u"Basic Latin"))
            c.add_class(CharClass.ucd_block(u"Latin-1 Supplement")"""
        global UCDBlocks
        if not UCDBlocks:
            # The block table is empty, so we need to load it
            load_block_table()
        return UCDBlocks[_normalize_block_name(block_name)]

    def __init__(self, *args):
        self.ranges = []
        self._clear_cache()
        for arg in args:
            if is_text(arg):
                # Each character in the string is put in the class
                for c in arg:
                    self.add_char(c)
            elif type(arg) in (tuple, list):
                self.add_range(arg[0], arg[1])
            elif isinstance(arg, CharClass):
                self.add_class(arg)
            else:
                raise ValueError

    def __repr__(self):
        result = ['CharClass(']
        first_range = True
        for r in self.ranges:
            if first_range:
                first_range = False
            else:
                result.append(', ')
            if r[0] == r[1]:
                result.append(repr(r[0]))
            else:
                result.append('(')
                result.append(repr(r[0]))
                result.append(',')
                result.append(repr(r[1]))
                result.append(')')
        result.append(')')
        return ''.join(result)

    _set_controls = {
        ul("\x07"): ul("\\a"),
        ul("\x08"): ul("\\b"),
        ul("\x09"): ul("\\t"),
        ul("\x0A"): ul("\\n"),
        ul("\x0B"): ul("\\v"),
        ul("\x0C"): ul("\\f"),
        ul("\x0D"): ul("\\r")}

    _set_escapes = ul("-\\]")
    _backslash = ul("\\")

    def _set_escape(self, c):
        """Escapes characters for inclusion in a set, i.e., -, \,  and ]"""
        if c in self._set_escapes:
            return self._backslash + c
        else:
            return self._set_controls.get(c, c)

    _re_controls = {
        ul("\x07"): ul("\\a"),
        ul("\x09"): ul("\\t"),
        ul("\x0A"): ul("\\n"),
        ul("\x0B"): ul("\\v"),
        ul("\x0C"): ul("\\f"),
        ul("\x0D"): ul("\\r")}

    _re_escapes = ul(".^$*+?{}\\[]|()")

    def _re_escape(self, c):
        """Escapes characters for inclusion in a regular expression
        outside a set"""
        if c in self._re_escapes:
            return self._backslash + c
        else:
            return self._re_controls.get(c, c)

    _empty_range = ul("[^\\x00-%s]") % character(maxunicode)

    def __unicode__(self):
        result = []
        if len(self.ranges) == 0:
            # we generally try and avoid representing maxunicode in a range
            # by using negation.  However, in the case of an empty range
            # we have no choice.
            return self._empty_range
        elif self.ranges[-1][1] == character(maxunicode):
            # to avoid maxunicode we negate this range
            neg = CharClass(self)
            neg.negate()
            result = to_text(neg)
            if result[0] == "[":
                return "[^%s]" % result[1:-1]
            elif result[0] == "\\":
                # we may not need the escape
                if result == "\\]":
                    return "[^\\]]"
                elif result in dict_values(self._re_controls):
                    return "[^%s]" % result
                else:
                    return "[^%s]" % result[1]
            else:
                return "[^%s]" % result
        if len(self.ranges) == 1:
            r = self.ranges[0]
            if r[0] == r[1]:
                # a single character
                return self._re_escape(r[0])
        add_caret = False
        for r in self.ranges:
            if r[0] == "^":
                add_caret = True
                r0 = u"_"
            else:
                r0 = r[0]
            if ord(r0) > ord(r[1]):
                continue
            elif r0 == r[1]:
                # just a singleton
                result.append(self._set_escape(r0))
            elif ord(r0) + 1 == ord(r[1]):
                # a dumb range, remove the hyphen
                result.append(
                    "%s%s" % (self._set_escape(r0), self._set_escape(r[1])))
            else:
                result.append("%s-%s" %
                              (self._set_escape(r0), self._set_escape(r[1])))
        if add_caret:
            result.append(u'^')
        return u"[%s]" % u"".join(result)

    def format_re(self):
        """Create a representation of the class suitable for putting in
        [] in a python regular expression"""
        py_charset = []
        for a, z in self.ranges:
            py_charset.append(self.format_re_char(a))
            if a == z:
                continue
            if ord(z) > ord(a) + 1:
                py_charset.append('-')
            py_charset.append(self.format_re_char(z))
        return ''.join(py_charset)

    def format_re_char(self, c):
        if c in "-]\\":
            # prepen a backslash
            return "\\" + c
        else:
            return c

    def __eq__(self, other):
        """Compares two character classes for equality."""
        return self.ranges == other.ranges

    def add_range(self, a, z):
        """Adds a range of characters from a to z to the class"""
        # our implementation assumes that codepoint is used in
        # comparisons
        a = force_text(a)
        z = force_text(z)
        if z < a:
            x = z
            z = a
            a = x
        if self.ranges:
            match_a, index_a = self._bisection_search(a, 0,
                                                      len(self.ranges) - 1)
            match_z, index_z = self._bisection_search(z, 0,
                                                      len(self.ranges) - 1)
            if match_a:
                if match_z:
                    # Both ends of the new range are already matched
                    if index_a == index_z:
                        # Nothing to do
                        return
                    else:
                        # We need to join the ranges from index_a to and
                        # including index_z
                        self.ranges[index_a:index_z + 1] = [
                            [self.ranges[index_a][0], self.ranges[index_z][1]]]
                else:
                    # Note that at this point, index_z must be > index_a
                    # We need to join the ranges from index_a up to but
                    # *not* including index_z extending the last range to
                    # include z
                    self.ranges[
                        index_a:index_z] = [[self.ranges[index_a][0], z]]
            elif match_z:
                # We need to join the ranges from index_a up to and
                # including index_z extending the first range to include
                # a (works even if index_a==index_z)
                self.ranges[
                    index_a:index_z + 1] = [[a, self.ranges[index_z][1]]]
            else:
                # We need to join the ranges from index_a to index_z-1,
                # extending them to include a and z respectively.  Note
                # that if index_a==index_z then no ranges are joined and
                # the slice assignment simply inserts a new range.
                self.ranges[index_a:index_z] = [[a, z]]
            self._merge(index_a)
        else:
            self.ranges = [[a, z]]
        self._clear_cache()

    def subtract_range(self, a, z):
        """Subtracts a range of characters from the character class"""
        a = force_text(a)
        z = force_text(z)
        if z < a:
            x = z
            z = a
            a = x
        if self.ranges:
            match_a, index_a = self._bisection_search(a, 0,
                                                      len(self.ranges) - 1)
            match_z, index_z = self._bisection_search(z, 0,
                                                      len(self.ranges) - 1)
            if match_a:
                if match_z:
                    # Both ends of the new range are matched
                    if index_a == index_z:
                        # a-z is entirely within a single range
                        rlower, rupper = self.ranges[index_a]
                        if ord(rlower) == ord(rupper) or (
                                ord(rlower) == ord(a) and
                                ord(rupper) == ord(z)):
                            # This is either a singleton range, so a==z
                            # must be true too! or we have an exact
                            # range match
                            del self.ranges[index_a]
                        elif ord(rlower) == ord(a):
                            # Remove the left portion of the range
                            self.ranges[index_a][0] = character(ord(z) + 1)
                        elif ord(rupper) == ord(z):
                            # Remove the right portion of the range
                            self.ranges[index_a][1] = character(ord(a) - 1)
                        else:
                            # We need to split this range
                            self.ranges[index_a][1] = character(ord(a) - 1)
                            self.ranges.insert(
                                index_a + 1,
                                [character(ord(z) + 1), rupper])
                    else:
                        # We need to trim index_a and index_z and remove all
                        # ranges between
                        rlower, rupper = self.ranges[index_a]
                        if ord(rlower) == ord(rupper) or ord(a) == ord(rlower):
                            # remove this entire range
                            snip_a = index_a
                        else:
                            # Remove the right portion of the range
                            self.ranges[index_a][1] = character(ord(a) - 1)
                            snip_a = index_a + 1
                        rlower, rupper = self.ranges[index_z]
                        if ord(rlower) == ord(rupper) or ord(z) == ord(rupper):
                            # remove this entire range
                            snip_z = index_z + 1
                        else:
                            # Remove the left portion of the range
                            self.ranges[index_z][0] = character(ord(z) + 1)
                            snip_z = index_z
                        if snip_z >= snip_a:
                            del self.ranges[snip_a:snip_z]
                else:
                    # We need to trim index_a and delete up to, but not
                    # including, index_z
                    rlower, rupper = self.ranges[index_a]
                    if ord(rlower) == ord(rupper) or ord(a) == ord(rlower):
                        snip = index_a
                    else:
                        self.ranges[index_a][1] = character(ord(a) - 1)
                        snip = index_a + 1
                    del self.ranges[snip:index_z]
            elif match_z:
                # We need to trim index_z and delete to the left up to and
                # including index_a
                rlower, rupper = self.ranges[index_z]
                if ord(rlower) == ord(rupper) or ord(z) == ord(rupper):
                    snip = index_z + 1
                else:
                    self.ranges[index_z][0] = character(ord(z) + 1)
                    snip = index_z
                del self.ranges[index_a:snip]
            else:
                # We need to remove the ranges from index_a to index_z-1.
                # Note that if index_a==index_z then no ranges are removed
                del self.ranges[index_a:index_z]
        self._clear_cache()

    def add_char(self, c):
        """Adds a single character to the character class"""
        c = force_text(c)
        if self.ranges:
            match, index = self._bisection_search(c, 0, len(self.ranges) - 1)
            if not match:
                self.ranges.insert(index, [c, c])
                self._merge(index)
        else:
            self.ranges = [[c, c]]
        self._clear_cache()

    def subtract_char(self, c):
        """Subtracts a single character from the character class"""
        c = force_text(c)
        if self.ranges:
            match, index = self._bisection_search(c, 0, len(self.ranges) - 1)
            if match:
                a, z = self.ranges[index]
                if ord(a) == ord(z):
                    # This is a singleton range
                    del self.ranges[index]
                elif ord(a) == ord(c):
                    self.ranges[index][0] = character(ord(a) + 1)
                elif ord(z) == ord(c):
                    self.ranges[index][1] = character(ord(z) - 1)
                else:
                    # We need to split this range
                    self.ranges[index][1] = character(ord(c) - 1)
                    self.ranges.insert(index + 1,
                                       [character(ord(c) + 1), z])
        self._clear_cache()

    def add_class(self, c):
        """Adds all the characters in c to the character class

        This is effectively a union operation."""
        if self.ranges:
            for r in c.ranges:
                self.add_range(r[0], r[1])
        else:
            # take a short cut here, if we have no ranges yet just copy them
            for r in c.ranges:
                self.ranges.append(r)
        self._clear_cache()

    def subtract_class(self, c):
        """Subtracts all the characters in c from the character class"""
        for r in c.ranges:
            self.subtract_range(r[0], r[1])
        self._clear_cache()

    def negate(self):
        """Negates this character class

        As a convenience returns the object as the result enabling
        this method to be used in construction, e.g.::

            c = CharClass('\x0a\x0d').negate()

        Results in the class of all characters *except* line feed and
        carriage return."""
        max = CharClass([character(0), character(maxunicode)])
        max.subtract_class(self)
        self.ranges = max.ranges
        self._clear_cache()
        return self

    def _merge(self, index):
        """Used internally to merge the range at index with its
        neighbours if possible"""
        a, z = self.ranges[index]
        index_a = index_z = index
        if index_a > 0:
            ap = self.ranges[index_a - 1][1]
            if ord(ap) >= ord(a) - 1:
                # Left merge
                index_a = index_a - 1
        if index_z < len(self.ranges) - 1:
            zn = self.ranges[index_z + 1][0]
            if ord(zn) <= ord(z) + 1:
                # Right merge
                index_z = index_z + 1
        if index_a != index_z:
            # Do the merge
            self.ranges[
                index_a:index_z + 1] = [[self.ranges[index_a][0],
                                         self.ranges[index_z][1]]]

    def _clear_cache(self):
        self._block_cache = [None] * 256

    def test(self, c):
        """Test a unicode character.

        Returns True if the character is in the class.

        If c is None, False is returned.

        This function uses an internal cache to speed up tests of
        complex classes.  Test results are cached in 256 character
        blocks.  The cache does not require a lock to make this method
        thread-safe (a lock would have a significant performance
        penalty) as it uses a simple python list.  The worst case race
        condition would result in two separate threads calculating the
        same block simultaneously and assigning it the same slot in the
        cache but python's list object is thread-safe under assignment
        (and the two calculated blocks will be identical) so this is not
        an issue.

        Why does this matter?  This function is called a lot,
        particularly when parsing XML.  When parsing a tag the parser
        will repeatedly test each character to determine if it is a
        valid name character and the definition of name character is
        complex.  Here are some illustrative figures calculated using
        cProfile for a typical 1MB XML file which calls test 142198
        times: with no cache 0.42s spent in test, with the cache 0.11s
        spent."""
        if c is None:
            return False
        elif self.ranges:
            cv = ord(c)
            block_num = cv >> 8
            try:
                block = self._block_cache[block_num]
                if block is None:
                    block = bytearray(256)
                    ccode = block_num * 256
                    match, index = self._bisection_search(
                        character(ccode), 0, len(self.ranges) - 1)
                    # match == True means we're in the indexed range
                    # match == False means indexed range is after us
                    block[0] = match
                    if index < len(self.ranges):
                        rmatch = self.ranges[index]
                        cmin = ord(rmatch[0])
                        cmax = ord(rmatch[1])
                    else:
                        rmatch = False
                    for i in range3(1, 256):
                        ccode += 1
                        if match:
                            # have we left this range?
                            if ccode > cmax:
                                index += 1
                                if index < len(self.ranges):
                                    rmatch = self.ranges[index]
                                    cmin = ord(rmatch[0])
                                    cmax = ord(rmatch[1])
                                else:
                                    rmatch = None
                                match = False
                            else:
                                block[i] = True
                                continue
                        # have we entered this range
                        if rmatch and ccode >= cmin:
                            block[i] = True
                            match = True
                    self._block_cache[block_num] = block
                block_pos = cv & 0xFF
                if block[block_pos]:
                    return True
                else:
                    return False
            except IndexError:
                # deal with wide Unicode characters using standard search
                match, index = self._bisection_search(
                    c, 0, len(self.ranges) - 1)
                return match
        else:
            return False

    def test_slow(self, c):
        if c is None:
            return False
        elif self.ranges:
            match, index = self._bisection_search(
                c, 0, len(self.ranges) - 1)
            return match
        else:
            return False

    def _bisection_search(self, c, rmin, rmax):
        """Performs a recursive bisection search for c.

        c is the character to search for rmin and rmax define a slice on
        the list of ranges in which to search

        The result is a tuple comprising a flag indicating if c is in
        the part of the class being searched and an integer index of the
        range into which c falls or, if c was not found, then it is the
        index at which a new range (containing only c) should be
        inserted."""
        if rmin == rmax:
            # is c in this range
            if c > self.ranges[rmin][1]:
                return (False, rmin + 1)
            elif c < self.ranges[rmin][0]:
                return (False, rmin)
            else:
                return (True, rmin)
        else:
            rtry = (rmin + rmax) // 2
            if c <= self.ranges[rtry][1]:
                return self._bisection_search(c, rmin, rtry)
            else:
                return self._bisection_search(c, rtry + 1, rmax)


def load_category_table():
    """Loads the category table from a resource file."""
    global UCDCategories
    f = open(os.path.join(os.path.dirname(__file__), CATEGORY_FILE), 'rb')
    UCDCategories = load(f)
    f.close()


def _get_cat_class(cat_name):
    global UCDCategories
    if cat_name in UCDCategories:
        return UCDCategories[cat_name]
    else:
        cat = CharClass()
        UCDCategories[cat_name] = cat
        return cat


def parse_category_table():
    global UCDCategories
    UCDCategories = {}
    next_code = 0
    mark = None
    mark_major_category = None
    mark_minor_category = None
    n_major_cat = _get_cat_class(u'C')
    n_minor_cat = _get_cat_class(u'Cn')
    for line in urlopen(UCDDatabaseURL).readlines():
        # disregard any comments
        line = to_text(line)
        line = line.split('#')[0]
        if not line:
            continue
        fields = line.split(';')
        code_point = int(fields[0], 16)
        assert code_point >= next_code, \
            "Unicode database error: code points went backwards: at %08X" % \
            code_point
        if code_point > maxunicode:
            logging.warning(
                "Warning: category table limited by narrow python build")
            break
        category = fields[2].strip()
        assert len(category) == 2, "Unexpected category field"
        major_category = _get_cat_class(category[0])
        minor_category = _get_cat_class(category)
        char_name = fields[1].strip()
        if mark is None:
            if char_name[0] == '<' and char_name[-6:] == "First>":
                mark = code_point
                mark_major_category = major_category
                mark_minor_category = minor_category
            else:
                major_category.add_char(character(code_point))
                minor_category.add_char(character(code_point))
            if code_point > next_code:
                # we have skipped a load of code-points
                n_major_cat.add_range(character(next_code),
                                      character(code_point - 1))
                n_minor_cat.add_range(character(next_code),
                                      character(code_point - 1))
        else:
            # end a marked range
            assert minor_category == mark_minor_category, \
                "Unicode character range end-points with non-matching "\
                "general categories"
            mark_major_category.add_range(character(mark),
                                          character(code_point))
            mark_minor_category.add_range(character(mark),
                                          character(code_point))
            mark = None
            mark_major_category = None
            mark_minor_category = None
        next_code = code_point + 1
    # when we finally exit from this loop we should not be in a marked range
    assert mark is None, \
        "Unicode database ended during character range definition: %08X-?" % \
        mark
    f = open(os.path.join(os.path.dirname(__file__), CATEGORY_FILE), 'wb')
    dump(UCDCategories, f)
    f.close()


def load_block_table():
    """Loads the block table from a resource file."""
    global UCDBlocks
    f = open(os.path.join(os.path.dirname(__file__), BLOCK_FILE), 'rb')
    UCDBlocks = load(f)
    f.close()


def _normalize_block_name(block_name):
    """Implements Unicode name normalization for block names.

    Removes white space, '-', '_' and forces lower case."""
    block_name = ''.join(block_name.split())
    block_name = block_name.replace('-', '')
    return block_name.replace('_', '').lower()


def parse_block_table():
    global UCDBlocks
    UCDBlocks = {}
    narrow_warning = False
    for line in urlopen(UCDBlockDatabaseURL).readlines():
        line = to_text(line)
        line = line.split('#')[0].strip()
        if not line:
            continue
        fields = line.split(';')
        code_points = fields[0].strip().split('..')
        code_point0 = int(code_points[0], 16)
        code_point1 = int(code_points[1], 16)
        # the Unicode standard tells us to remove -, _ and any whitespace
        # before case-ignore comparison
        block_name = _normalize_block_name(fields[1])
        if code_point0 > maxunicode:
            if not narrow_warning:
                logging.warning(
                    "Warning: block table limited by narrow python build")
            narrow_warning = True
            continue
        elif code_point1 > maxunicode:
            code_point1 = maxunicode
            if not narrow_warning:
                logging.warning(
                    "Warning: block table limited by narrow python build")
            narrow_warning = True
        UCDBlocks[block_name] = CharClass(
            (character(code_point0), character(code_point1)))
    f = open(os.path.join(os.path.dirname(__file__), BLOCK_FILE), 'wb')
    dump(UCDBlocks, f)
    f.close()


class ParserError(ValueError):

    """Exception raised by :class:`BasicParser`

    production
        The name of the production being parsed

    parser
        The :class:`BasicParser` instance raising the error (optional)

    ParserError is a subclass of ValueError."""

    def __init__(self, production, parser=None):
        self.production = production
        if parser:
            #: the position of the parser when the error was raised
            self.pos = parser.pos
            #: up to 40 characters/bytes to the left of pos
            self.left = parser.src[max(0, self.pos - 40):self.pos]
            #: up to 40 characters/bytes to the right of pos
            self.right = parser.src[self.pos:self.pos + 40]
            if production:
                msg = "ParserError: expected %s at [%i]" % (production,
                                                            self.pos)
            else:
                msg = "ParserError: at [%i]" % self.pos
        else:
            self.pos = None
            self.left = None
            self.right = None
            if production:
                msg = "ParserError: expected %s" % production
            else:
                msg = "ParserError"
        ValueError.__init__(self, msg)


class ParserMixin(object):

    """A mix-in class for parsing

    These methods help establish a common pattern across parsers by
    providing a conversion between look-ahead and non-look ahead style
    parsing methods.

    Derived classes must define the following.

        parser_error
            a method for raising an error detected during parsing that
            takes an optional character string describing the production
            being parsed.

        match_end
            a method for determining if the parser has consumed all
            the input.

        pos
            an attribute containing the current position

        setpos
            a method for setting *pos* from a previously saved value."""

    def require_production(self, result, production=None):
        """Returns *result* if not None or raises ParserError.

            result
                The result of a parse_* type method.

            production
                Optional string used to customise the error message.

        This method is intended to be used as a conversion function
        allowing any parse_* method to be converted into a require_*
        method.  E.g.::

            p = BasicParser("hello")
            num = p.require_production(p.parse_integer(), "Number")

            ParserError: Expected Number at [0]"""
        if result is None:
            self.parser_error(production)
        else:
            return result

    def parse_production(self, require_method, *args, **kwargs):
        """Executes the bound method *require_method*.

            require_method
                A bound method that will be called with \*args

            args
                The positional arguments to pass to require_method

            kwargs
                The keyword arguments to pass to require_method

        This method is intended to be used as a conversion function
        allowing any require_* method to be converted into a parse_*
        method for the purposes of look-ahead.

        If successful the result of the method is returned.  If any
        ValueError (including :class:`ParserError`) is raised, the
        exception is caught, the parser rewound and None is returned."""
        savepos = self.pos
        try:
            return require_method(*args, **kwargs)
        except ValueError:
            self.setpos(savepos)
            return None

    def require_end(self, production='end'):
        """Tests that the parser has consumed all the source

        There is no return result."""
        if not self.match_end():
            self.parser_error(production)

    def require_production_end(self, result, production=None):
        """Returns *result* if not None and parsing is complete.

        This method is similar to :meth:`require_production` except that
        it enforces the constraint that the entire source must have been
        parsed.  Essentially, it just calls :meth:`require_end` before
        returning *result*."""
        result = self.require_production(result, production)
        self.require_end(production)
        return result


class BasicParser(ParserMixin, PEP8Compatibility):

    """A base class for parsing character strings or binary data

    source
        Can be either a string of characters or a string of bytes.

    BasicParser instances can parse either characters or bytes but not
    both simultaneously, you must choose on construction by passing an
    appropriate str (Python 2: unicode), bytes or bytearray object.

    Binary mode is suitable for parsing data described in terms of
    OCTETS, such as many IETF and internet standards.  When passing
    string literals to parsing methods in binary mode use the binary
    string literal form::

        parser.match(b':')

    Methods that return the parsed data in its original form will also
    return bytes objects in binary mode.

    Methods are named according to the type of operation they perform.

        match\_*
            Returns a boolean True or False depending on whether or not
            a syntax production is matched at the current location. The
            state of the parser is unchanged.  This type of method is
            only used for very simple productions, e.g.,
            :meth:`match_digit`.

        parse\_*
            Attempts to parse a syntax element returning an appropriate
            object as the result or None if the production is not
            present. The position of the parser is only changed if the
            element was parsed successfully.  This type of method is
            intended for fairly simple productions, e.g.,
            :meth:`parse_integer`. More complex productions are
            implemented using require\_* methods but the general
            :meth:`parse_production` can be used to enable more complex
            look-ahead scenarios.

        require\_*
            Parses a syntax production, returning an appropriate object
            as the result.  If the production is not matched a
            :class:`ParserError` is raised.

            On success, the position of the parser points to the first
            character after the parsed production ready to continue
            parsing.  On failure, the parser is positioned at the
            point at which the exception was raised.

            When deriving your own sub-classes you will normally use the
            require\_* pattern to extend the parser.

    Compatibility note: if you are attempting to use the same source for
    both Python 2 and 3 then you may not be able to rely on the parser
    mode::

        >>> from pyslet.unicode5 import BasicParser
        >>> p = BasicParser("hello")
        >>> p.raw

    The above interpreter session will print True in Python 2 and False
    in Python 3.  This is just another manifestation of the changes to
    string handling between the two releases.  If you are dealing with
    ASCII data you can ignore the issue, otherwise you should consider
    using one of the various techniques for forcing strings to be
    interpreted as unicode when running in Python 2.  The most important
    thing is consistency between the type of object you pass to the
    constructor and those that you pass to the various parsing
    methods.  You may find the :func:`pyslet.py2.ul` and/or
    :func:`pyslet.py2.u8` functions useful for forcing text mode."""

    def __init__(self, source):
        PEP8Compatibility.__init__(self)
        self.raw = not is_unicode(source)
        """True if parser is working in binary mode."""
        self.src = source       #: the string being parsed
        self.pos = -1           #: the position of the current character
        self.the_char = None
        """The current character or None if the parser is positioned
        outside the src string.

        In binary mode this will be a byte, which is an integer in
        Python 3 but a character in Python 2.  In text mode it is
        a (unicode) character."""
        self.last_error = None
        self.next_char()

    def setpos(self, new_pos):
        """Sets the position of the parser to *new_pos*

        Useful for saving the parser state and returning later::

            save_pos = parser.pos
            #
            # do some look-ahead parsing
            #
            parser.setpos(save_pos)
        """
        self.pos = new_pos - 1
        self.next_char()

    def next_char(self):
        """Points the parser at the next character.

        Updates *pos* and *the_char*."""
        self.pos += 1
        if self.pos >= 0 and self.pos < len(self.src):
            self.the_char = self.src[self.pos]
        else:
            self.the_char = None

    def parser_error(self, production=None):
        """Raises an error encountered by the parser

        See :class:`ParserError` for details.

        If production is None then the previous error is re-raised. If
        multiple errors have been raised previously the one with the
        most advanced parser position is used.  This is useful in
        situations where there are multiple alternative productions,
        none of which can be successfully parsed.  It allows parser
        methods to catch the exception from the last possible choice and
        raise an error relating to the closest previous match.  For
        example::

            def require_abc(self):
                result = p.parse_production(p.require_a)
                if result is None:
                    result = p.parse_production(p.require_b)
                if result is None:
                    result = p.parse_production(p.require_c)
                if result is None:
                    # will raise the most advanced error raised during
                    # the three previous methods
                    p.parser_error()
                else:
                    return result

        See :meth:`parse_production` for more details on this pattern.

        The position of the parser is always set to the position of the
        error raised."""
        if production:
            e = ParserError(production, self)
        elif self.last_error is not None and self.pos <= self.last_error.pos:
            e = self.last_error
        else:
            e = ParserError('', self)
        if self.last_error is None or e.pos > self.last_error.pos:
            self.last_error = e
        if e.pos != self.pos:
            self.setpos(e.pos)
        raise e

    def peek(self, nchars):
        """Returns the next *nchars* characters or bytes.

        If there are less than nchars remaining then a shorter string is
        returned."""
        return self.src[self.pos:self.pos + nchars]

    def match_end(self):
        """True if all of :attr:`src` has been parsed"""
        return self.the_char is None

    def match(self, match_string):
        """Returns true if *match_string* is at the current position"""
        if self.the_char is None:
            return False
        else:
            return self.src[self.pos:self.pos +
                            len(match_string)] == match_string

    def parse(self, match_string):
        """Parses *match_string*

        Returns *match_string* or None if it cannot be parsed."""
        if self.match(match_string):
            self.setpos(self.pos + len(match_string))
            return match_string
        else:
            return None

    def require(self, match_string, production=None):
        """Parses and requires *match_string*

        match_string
            The string to be parsed

        production
            Optional name of production, defaults to match_string itself.

        For consistency, returns match_string on success."""
        if not self.parse(match_string):
            if production is None:
                if self.raw:
                    production = repr(match_string)
                    if py2 and production and production[0] in "\"'":
                        production = "b" + production
                else:
                    production = match_string

            self.parser_error(production)
        else:
            return match_string

    def match_insensitive(self, lower_string):
        """Returns true if *lower_string* is matched (ignoring case).

        *lower_string* must already be a lower-cased string."""
        if self.the_char is None:
            return False
        else:
            return self.src[self.pos:self.pos +
                            len(lower_string)].lower() == lower_string

    def parse_insensitive(self, lower_string):
        """Parses *lower_string* ignoring case in the source.

        lower_string
            Must be a lower-cased string

        Advances the parser to the first character after lower_string.
        Returns the matched string which may differ in case from
        lower_string."""
        if self.match_insensitive(lower_string):
            return_string = self.src[self.pos:self.pos + len(lower_string)]
            self.setpos(self.pos + len(lower_string))
            return return_string
        else:
            return None

    def parse_until(self, match_string):
        """Parses up to but not including *match_string*.

        Advances the parser to the first character *of* match_string.
        If match_string is not found (or is None) then all the remaining
        characters in the source are parsed.

        Returns the parsed text, even if empty.  Never returns None."""
        if match_string is None:
            match_pos = -1
        else:
            match_pos = self.src.find(match_string, self.pos)
        if match_pos == -1:
            result = self.src[self.pos:]
            self.setpos(len(self.src))
        else:
            result = self.src[self.pos:match_pos]
            self.setpos(match_pos)
        return result

    def match_one(self, match_chars):
        """Returns true if one of *match_chars* is at the current position.

        The 'in' operator is used to test match_chars so this can be a
        list or tuple of characters (or bytes), it does not have to be
        string."""
        if self.the_char is None:
            return False
        else:
            return self.the_char in match_chars

    def parse_one(self, match_chars):
        """Parses one of *match_chars*.

        match_chars
            A *string* (list or tuple) of characters or bytes

        Returns the character (or byte) or None if no match is found.

        Warning: in binary mode, this method will return a single byte
        value, the type of which will differ in Python 2.  In Python 3,
        bytes are integers, in Python 2 they are binary strings of
        length 1.  You can use the function :func:`py2.byte` to help
        ensure your source works on both platforms, for example::

            from .py2 import byte
            c = parser.parse_one(b"+-")
            if c == byte(b"+"):
                # do plus thing...
            elif c is not None:
                # must be minus...
            else:
                # do something else...
        """
        if self.match_one(match_chars):
            result = self.the_char
            self.next_char()
            return result
        else:
            return None

    u_digits = ul("0123456789")
    b_digits = b"0123456789"

    def match_digit(self):
        """Returns true if the current character is a digit

        Only ASCII digits are considered, in binary mode byte values
        0x30 to 0x39 are matched."""
        if self.raw:
            return self.match_one(self.b_digits)
        else:
            return self.match_one(self.u_digits)

    def parse_digit(self):
        """Parses a digit character.

        Returns the digit character/byte, or None if no digit is found.
        Like :meth:`match_digit` only ASCII digits are parsed."""
        if self.raw:
            return self.parse_one(self.b_digits)
        else:
            return self.parse_one(self.u_digits)

    def parse_digit_value(self):
        """Parses a single digit value.

        Returns the digit value, or None if no digit is found.
        Like :meth:`match_digit` only ASCII digits are parsed."""
        if self.raw:
            result = self.parse_one(self.b_digits)
            if result is not None:
                result = byte_value(result) - 0x30
        else:
            result = self.parse_one(self.u_digits)
            if result is not None:
                result = ord(result) - 0x30
        return result

    def parse_digits(self, min, max=None):
        """Parses a string of digits

        min
            The minimum number of digits to parse.  There is a special
            cases where min=0, in this case an empty string may be
            returned.

        max (default None)
            The maximum number of digits to parse, or None there is no
            maximum.

        Returns the string of digits or None if no digits can be parsed.
        Like :meth:`parse_digit`, only ASCII digits are considered."""
        if min < 0 or (max is not None and min > max):
            raise ValueError("min must be > 0")
        savepos = self.pos
        result = []
        while max is None or len(result) < max:
            d = self.parse_digit()
            if d is None:
                break
            else:
                result.append(d)
        if len(result) < min:
            self.setpos(savepos)
            return None
        if self.raw:
            return join_bytes(result)
        else:
            return ul('').join(result)

    def parse_integer(self, min=None, max=None, max_digits=None):
        """Parses an integer (or long).

        min (optional, defaults to None)
            A lower bound on the acceptable integer value, the
            result will always be >= min on success

        max (optional, defaults to None)
            An upper bound on the acceptable integer value, the
            result will always be <= max on success

        max_digits (optional, defaults to None)
            The limit on the number of digits, i.e., the field width.

        If a suitable integer can't be parsed then None is returned.
        This method only processes ASCII digits.

        Warning: in Python 2 the result may be of type long."""
        if min is None:
            min = 0
        if min < 0 or (max is not None and max < min):
            raise ValueError("0 <= min <= max required")
        savepos = self.pos
        d = self.parse_digits(1, max_digits)
        if d is None:
            return None
        else:
            d = int(d)
            if d < min or (max is not None and d > max):
                self.setpos(savepos)
                return None
            return d

    u_hex_digits = ul("0123456789abcdefABCDEF")
    b_hex_digits = b"0123456789abcdefABCDEF"

    def match_hex_digit(self):
        """Returns true if the current character is a hex-digit

        Only ASCII digits are considered, letters can be either upper or
        lower case.  In binary mode byte values 0x30 to 0x39, 0x41-0x46
        and 0x61-0x66 are matched."""
        if self.raw:
            return self.match_one(self.b_hex_digits)
        else:
            return self.match_one(self.u_hex_digits)

    def parse_hex_digit(self):
        """Parses a hex-digit.

        Returns the digit, or None if no digit is found.  See
        :meth:`match_hex_digit` for which characters/bytes are
        considered hex-digits."""
        if self.raw:
            return self.parse_one(self.b_hex_digits)
        else:
            return self.parse_one(self.u_hex_digits)

    def parse_hex_digits(self, min, max=None):
        """Parses a string of hex-digits

        min
            The minimum number of hex-digits to parse.  There is a
            special cases where min=0, in this case an empty string may
            be returned.

        max (default None)
            The maximum number of hex-digits to parse, or None there is
            no maximum.

        Returns the string of hex-digits or None if no digits can be
        parsed. See :meth:`match_hex_digit` for which characters/bytes
        are considered hex-digits."""
        if min < 0 or (max is not None and min > max):
            raise ValueError("min must be > 0")
        savepos = self.pos
        rlen = 0
        while max is None or rlen < max:
            d = self.parse_hex_digit()
            if d is None:
                break
            else:
                rlen += 1
        if rlen < min:
            self.setpos(savepos)
            return None
        return self.src[savepos:savepos + rlen]



#! /usr/bin/env python

import logging
import itertools

from . import core
from .. import html401 as html
from .. import imsmdv1p2p1 as imsmd
from ..http import messages
from ..pep8 import old_method
from ..py2 import is_text, to_text
from ..qtiv2 import xml as qtiv2
from ..xml import structures as xml
from ..xml import xsdatatypes as xsi


class QTIUnimplementedOperator(core.QTIUnimplementedError):
    pass


class QTICommentContainer(core.QTIElement):

    """Basic element to represent all elements that can contain a comment
    as their first child::

            <!ELEMENT XXXXXXXXXXXX (qticomment? , ....... )>"""

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.QTIComment = None

    def get_children(self):
        if self.QTIComment:
            yield self.QTIComment


class QTIComment(core.QTIElement):

    """This element contains the comments that are relevant to the host element.
    The comment is contained as a string::

            <!ELEMENT qticomment (#PCDATA)>
            <!ATTLIST qticomment  xml:lang CDATA  #IMPLIED >"""
    XMLNAME = 'qticomment'
    XMLCONTENT = xml.XMLMixedContent


class Duration(core.QTIElement):

    """The duration permitted for the completion of a particular activity. The
    duration is defined as per the ISO8601 standard. The information is entered
    as a string::

            <!ELEMENT duration (#PCDATA)>"""
    XMLNAME = "duration"
    XMLCONTENT = xml.XMLMixedContent


class ContentMixin(object):

    """Mixin class for handling all content-containing elements.

    This class is used by all elements that behave as content, the
    default implementation provides an additional contentChildren member
    that should be used to collect any content-like children."""

    def __init__(self):
        self.contentChildren = []		#: the list of content children

    def content_child(self, child_class):
        """Returns True if child_class is an allowed subclass of
        ContentMixin in this context."""
        return True

    def add_child(self, child_class, name=None):
        """Creates a new child of this element.

        Overrides the underlying Element class to implement special
        handling for ContentMixin and its subclasses.  By default we
        accept any type of content but derived classes override this
        behaviour by providing an implementation of content_child to
        limit the range of elements to match their own content models."""
        if issubclass(child_class, ContentMixin):
            if self.content_child(child_class):
                child = child_class(self)
                self.contentChildren.append(child)
                if name:
                    child.set_xmlname(name)
                return child
            else:
                raise TypeError
        else:
            return super(ContentMixin, self).add_child(child_class, name)

    @old_method('GetContentChildren')
    def get_content_children(self):
        """Returns an iterable of the content children."""
        return iter(self.contentChildren)

    @old_method('IsInline')
    def is_inline(self):
        """True if this element can be inlined, False if it is block level

        The default implementation returns True if all
        :py:attr:`contentChildren` can be inlined."""
        return self.inline_children()

    @old_method('InlineChildren')
    def inline_children(self):
        """True if all of this element's :py:attr:`contentChildren` can all be
        inlined."""
        for child in self.contentChildren:
            if not child.is_inline():
                return False
        return True

    @old_method('ExtractText')
    def extract_text(self):
        """Returns a tuple of (<text string>, <lang>).

        Sometimes it is desirable to have a plain text representation of a
        content object.  For example, an element may permit arbitrary content
        but a synopsis is required to set a metadata value.

        Our algorithm for determining the language of the text is to first
        check if the language has been specified for the context.  If it has
        then that language is used.  Otherwise the first language attribute
        encountered in the content is used as the language.  If no language is
        found then None is returned as the second value."""
        result = []
        lang = self.get_lang()
        for child in self.contentChildren:
            childText, childLang = child.extract_text()
            if lang is None:
                lang = childLang
            if childText:
                result.append(childText.strip())
        return ' '.join(result), lang

    @old_method('MigrateV2Content')
    def migrate_content_to_v2(self, parent, child_type, log, children=None):
        """Migrates this content element to QTIv2.

        The resulting QTIv2 content is added to *parent*.

        *child_type* indicates whether the context allows block, inline or a
        mixture of element content types (flow).  It is set to one of the
        following HTML classes: :py:class:`pyslet.html401.BlockMixin`,
        :py:class:`pyslet.html401.InlineMixin` or
        :py:class:`pyslet.html401.FlowMixin`.

        The default implementation adds each of *children* or, if *children* is
        None, each of the local :py:attr:`contentChildren`.  The algorithm
        handles flow elements by creating <p> elements where the context
        permits.  Nested flows are handled by the addition of <br/>."""
        if children is None:
            children = self.contentChildren
        if child_type is html.InlineMixin or (child_type is html.FlowMixin and
                                              self.inline_children()):
            # We can only hold inline children, raise an error if find anything
            # else
            br_before = br_after = False
            first_item = True
            for child in children:
                if isinstance(child, FlowMixin):
                    br_before = not first_item
                    br_after = True
                elif br_after:
                    # we only honour br_after if flow is followed by something
                    # other than flow
                    parent.add_child(html.Br,
                                     (qtiv2.core.IMSQTI_NAMESPACE, 'br'))
                    br_after = False
                if br_before:
                    parent.add_child(
                        html.Br, (qtiv2.core.IMSQTI_NAMESPACE, 'br'))
                    br_before = False
                # we force InlineMixin here to prevent unnecessary checks on
                # inline status
                child.migrate_content_to_v2(parent, html.InlineMixin, log)
                first_item = False
        else:
            # child_type is html.BlockMixin or html.FlowMixin and we have a
            # mixture of inline/block children
            p = None
            br_before = False
            br_after = False
            for child in children:
                try:
                    if child.is_inline():
                        if br_after:
                            p.add_child(
                                html.Br, (qtiv2.core.IMSQTI_NAMESPACE, 'br'))
                            br_after = False
                        if isinstance(child, FlowMixin):
                            br_before = br_after = True
                        if p is None:
                            p = parent.add_child(
                                html.P, (qtiv2.core.IMSQTI_NAMESPACE, 'p'))
                            br_before = False
                        if br_before:
                            p.add_child(html.Br)
                            br_before = False
                        child.migrate_content_to_v2(p, html.InlineMixin, log)
                    else:
                        # stop collecting inlines
                        p = None
                        br_before = br_after = False
                        child.migrate_content_to_v2(parent, html.BlockMixin,
                                                    log)
                except AttributeError as e:
                    logging.error("migrate_content_to_v2: %s", str(e))
                    raise core.QTIError(
                        "Error: unsupported QTI v1 content element " +
                        child.xmlname)


class MatThingMixin(ContentMixin):

    """An abstract class used to help identify the mat* elements."""
    pass


class Material(ContentMixin, QTICommentContainer):

    """This is the container for any content that is to be displayed by the
    question-engine. The supported content types are text (emphasized or not),
    images, audio, video, application and applet. The content can be internally
    referenced to avoid the need for duplicate copies. Alternative information
    can be defined - this is used if the primary content cannot be displayed::

            <!ELEMENT material (qticomment? , (mattext | matemtext | matimage |
                    mataudio | matvideo | matapplet | matapplication | matref |
                    matbreak | mat_extension)+ , altmaterial*)>
            <!ATTLIST material
                    label CDATA  #IMPLIED
                    xml:lang CDATA  #IMPLIED >"""
    XMLNAME = 'material'
    XMLATTR_label = 'label'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        QTICommentContainer.__init__(self, parent)
        ContentMixin.__init__(self)
        self.AltMaterial = []
        self.label = None

    def content_child(self, child_class):
        return issubclass(child_class, MatThingMixin)

    def get_children(self):
        return itertools.chain(
            QTICommentContainer.get_children(self),
            ContentMixin.get_content_children(self))

    def content_changed(self):
        if self.label:
            doc = self.get_document()
            if doc:
                doc.register_material(self)
        QTICommentContainer.content_changed(self)


class AltMaterial(ContentMixin, QTICommentContainer):

    """This is the container for alternative content. This content is to be
    displayed if, for whatever reason, the primary content cannot be rendered.
    Alternative language implementations of the host <material> element are
    also supported using this structure::

            <!ELEMENT altmaterial (qticomment? ,
                    (mattext | matemtext | matimage | mataudio | matvideo |
                    matapplet | matapplication | matref | matbreak |
                    mat_extension)+)>
            <!ATTLIST altmaterial  xml:lang CDATA  #IMPLIED >"""
    XMLNAME = "altmaterial"
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        QTICommentContainer.__init__(self, parent)
        ContentMixin.__init__(self)

    def content_child(self, child_class):
        return issubclass(child_class, MatThingMixin)

    def get_children(self):
        return itertools.chain(
            QTICommentContainer.get_children(self),
            ContentMixin.get_content_children(self))


class PositionMixin:

    """Mixin to define the positional attributes
    ::

            width	CDATA  #IMPLIED
            height	CDATA  #IMPLIED
            y0		CDATA  #IMPLIED
            x0		CDATA  #IMPLIED"""
    XMLATTR_height = ('height', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_width = ('width', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_x0 = ('x0', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_y0 = ('y0', xsi.integer_from_str, xsi.integer_to_str)

    def __init__(self):
        self.x0 = None
        self.y0 = None
        self.width = None
        self.height = None

    def got_position(self):
        return (self.x0 is not None or self.y0 is not None or
                self.width is not None or self.height is not None)


class MatText(core.QTIElement, PositionMixin, MatThingMixin):

    """The <mattext> element contains any text that is to be displayed to the users
    ::

            <!ELEMENT mattext (#PCDATA)>
            <!ATTLIST mattext
                    texttype    CDATA  'text/plain'
                    label		CDATA  #IMPLIED
                    charset		CDATA  'ascii-us'
                    uri			CDATA  #IMPLIED
                    xml:space	(preserve | default )  'default'
                    xml:lang	CDATA  #IMPLIED
                    entityref	ENTITY  #IMPLIED
                    width		CDATA  #IMPLIED
                    height		CDATA  #IMPLIED
                    y0			CDATA  #IMPLIED
                    x0			CDATA  #IMPLIED >"""
    XMLNAME = 'mattext'
    XMLATTR_label = 'label'
    XMLATTR_charset = 'charset'
    XMLATTR_uri = ('uri', html.uri.URI.from_octets, html.to_text)
    XMLATTR_texttype = 'texttype'
    XMLCONTENT = xml.XMLMixedContent

    SymbolCharsets = {
        'greek': True,
        'symbol': True
    }

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        PositionMixin.__init__(self)
        MatThingMixin.__init__(self)
        self.texttype = 'text/plain'
        self.label = None
        self.charset = 'ascii-us'
        self.uri = None
        self.matChildren = []
        #: an inline html object used to wrap inline elements
        self.inlineWrapper = None

    def content_changed(self):
        if self.label:
            doc = self.get_document()
            if doc:
                doc.register_mat_thing(self)
        if self.texttype == 'text/html':
            if self.uri:
                # The content is external, load it up
                uri = self.resolve_uri(self.uri)
                try:
                    e = xml.XMLEntity(uri)
                except messages.HTTPException as e:
                    e = xml.XMLEntity(to_text(e))
            else:
                uri = self.resolve_base()
                try:
                    value = self.get_value()
                except xml.XMLMixedContentError:
                    # Assume that the element content was not protected by
                    # CDATA
                    value = []
                    for child in self.get_children():
                        value.append(to_text(child))
                    value = ''.join(value)
                if value:
                    e = xml.XMLEntity(value)
            doc = html.XHTMLDocument(baseURI=uri)
            doc.read_from_entity(e)
            self.matChildren = list(doc.root.Body.get_children())
            if (len(self.matChildren) == 1 and
                    isinstance(self.matChildren[0], html.Div)):
                div = self.matChildren[0]
                if div.style_class is None:
                    # a single div with no style class is removed...
                    self.matChildren = list(div.get_children())
        elif self.texttype == 'text/rtf':
            # parse the RTF content
            raise core.QTIUnimplementedError

    def is_inline(self):
        if self.texttype == 'text/plain':
            return True
        elif self.texttype == 'text/html':
            # we are inline if all elements in matChildren are inline
            for child in self.matChildren:
                if (is_text(child) or isinstance(child, html.InlineMixin)):
                    continue
                else:
                    return False
            return True
        else:
            raise core.QTIUnimplementedError(self.texttype)

    def extract_text(self):
        if self.matChildren:
            # we need to extract the text from these children
            results = []
            para = []
            for child in self.matChildren:
                if is_text(child):
                    para.append(child)
                elif isinstance(child, html.InlineMixin):
                    para.append(child.plain_text())
                else:
                    if para:
                        results.append(''.join(para))
                        para = []
                    results.append(child.plain_text())
            if para:
                results.append(''.join(para))
            return '\n\n'.join(results), self.resolve_lang()
        else:
            return self.get_value(), self.resolve_lang()

    def migrate_content_to_v2(self, parent, child_type, log):
        if self.texttype == 'text/plain':
            data = self.get_value(True)
            if self.charset.lower() in MatText.SymbolCharsets:
                # this data is in the symbol font, translate it
                # Note that the first page of unicode is the same as iso-8859-1
                data = data.encode('iso-8859-1').decode('apple-symbol')
            lang = self.get_lang()
            if lang or self.label:
                if child_type is html.BlockMixin:
                    span = parent.add_child(
                        html.P, (qtiv2.core.IMSQTI_NAMESPACE, 'p'))
                    if self.inlineWrapper:
                        span = span.add_child(
                            self.inlineWrapper,
                            (qtiv2.core.IMSQTI_NAMESPACE,
                             self.inlineWrapper.XMLNAME[1]))
                elif self.inlineWrapper:
                    span = parent.add_child(
                        self.inlineWrapper,
                        (qtiv2.core.IMSQTI_NAMESPACE,
                         self.inlineWrapper.XMLNAME[1]))
                else:
                    span = parent.add_child(
                        html.Span, (qtiv2.core.IMSQTI_NAMESPACE, 'span'))
                if lang:
                    span.set_lang(lang)
                if self.label:
                    # span.label=self.label
                    span.set_attribute('label', self.label)
                # force child elements to render as inline XML
                span.add_data(data)
            elif child_type is html.BlockMixin:
                p = parent.add_child(
                    html.P, (qtiv2.core.IMSQTI_NAMESPACE, 'p'))
                if self.inlineWrapper:
                    p = p.add_child(
                        self.inlineWrapper,
                        (qtiv2.core.IMSQTI_NAMESPACE,
                         self.inlineWrapper.XMLNAME[1]))
                p.add_data(data)
            else:
                # inline or flow, just add the text directly...
                if self.inlineWrapper:
                    add_node = parent.add_child(
                        self.inlineWrapper,
                        (qtiv2.core.IMSQTI_NAMESPACE,
                         self.inlineWrapper.XMLNAME[1]))
                else:
                    add_node = parent
                parent.add_data(data)
        elif self.texttype == 'text/html':
            if (child_type is html.BlockMixin or
                    (child_type is html.FlowMixin and not self.is_inline())):
                # Block or mixed-up flow, wrap all text and inline elements in
                # p
                p = None
                for child in self.matChildren:
                    if (is_text(child) or
                            isinstance(child, html.InlineMixin)):
                        if p is None:
                            p = parent.add_child(
                                html.P, (qtiv2.core.IMSQTI_NAMESPACE, 'p'))
                            if self.inlineWrapper:
                                p = p.add_child(
                                    self.inlineWrapper,
                                    (qtiv2.core.IMSQTI_NAMESPACE,
                                     self.inlineWrapper.XMLNAME[1]))
                        if is_text(child):
                            p.add_data(child)
                        else:
                            new_child = child.deepcopy(p)
                            qtiv2.content.fix_html_namespace(new_child)
                    else:
                        # stop collecting inlines
                        p = None
                        if self.inlineWrapper:
                            log.append(
                                'Warning: block level elements in text/html'
                                'cannot be wrapped with <%s>' %
                                self.inlineWrapper.XMLNAME[1])
                        new_child = child.deepcopy(parent)
                        qtiv2.content.fix_html_namespace(new_child)
            else:
                # Flow context (with only inline children) or inline context
                if self.inlineWrapper:
                    add_node = parent.add_child(
                        self.inlineWrapper,
                        (qtiv2.core.IMSQTI_NAMESPACE,
                         self.inlineWrapper.XMLNAME[1]))
                else:
                    add_node = parent
                for child in self.matChildren:
                    if is_text(child):
                        add_node.add_data(child)
                    else:
                        new_child = child.deepcopy(add_node)
                        qtiv2.content.fix_html_namespace(new_child)
        else:
            raise core.QTIUnimplementedError


class MatEmText(MatText):

    """The <matemtext> element contains any emphasized text that is to be
    displayed to the users. The type of emphasis is dependent on the
    question-engine rendering the text::

            <!ELEMENT matemtext (#PCDATA)>
            <!ATTLIST matemtext
                    texttype	CDATA  'text/plain'
                    label		CDATA  #IMPLIED
                    charset		CDATA  'ascii-us'
                    uri			CDATA  #IMPLIED
                    xml:space	(preserve | default )  'default'
                    xml:lang	CDATA  #IMPLIED
                    entityref	ENTITY  #IMPLIED
                    width		CDATA  #IMPLIED
                    height		CDATA  #IMPLIED
                    y0			CDATA  #IMPLIED
                    x0			CDATA  #IMPLIED >"""
    XMLNAME = "matemtext"
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        MatText.__init__(self, parent)
        self.inlineWrapper = html.Em


class MatBreak(core.QTIElement, MatThingMixin):

    """The element that is used to insert a break in the flow of the associated
    material. The nature of the 'break' is dependent on the display-rendering
    engine::

            <!ELEMENT matbreak EMPTY>"""
    XMLNAME = "matbreak"
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        MatThingMixin.__init__(self)

    def is_inline(self):
        return True

    def extract_text(self):
        """Returns a simple line break"""
        return "\n"

    def migrate_content_to_v2(self, parent, child_type, log):
        if child_type in (html.InlineMixin, html.FlowMixin):
            parent.add_child(html.Br, (qtiv2.core.IMSQTI_NAMESPACE, 'br'))
        else:
            # a break in a group of block level elements is ignored
            pass


class MatImage(core.QTIElement, PositionMixin, MatThingMixin):

    """The <matimage> element is used to contain image content that is to be
    displayed to the users::

            <!ELEMENT matimage (#PCDATA)>
            <!ATTLIST matimage
                    imagtype    CDATA  'image/jpeg'
                    label       CDATA  #IMPLIED
                    height      CDATA  #IMPLIED
                    uri         CDATA  #IMPLIED
                    embedded    CDATA  'base64'
                    width       CDATA  #IMPLIED
                    y0          CDATA  #IMPLIED
                    x0          CDATA  #IMPLIED
                    entityref   ENTITY #IMPLIED >"""
    XMLNAME = "matimage"
    XMLATTR_imagtype = 'imageType'
    XMLATTR_label = 'label'
    XMLATTR_uri = ('uri', html.uri.URI.from_octets, html.to_text)
    XMLATTR_embedded = 'embedded'
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        PositionMixin.__init__(self)
        MatThingMixin.__init__(self)
        self.imageType = 'image/jpeg'
        self.label = None
        self.uri = None
        self.embedded = 'base64'

    def is_inline(self):
        return True

    def extract_text(self):
        """We cannot extract text from matimage so we return a simple
        string."""
        return "[image]"

    def migrate_content_to_v2(self, parent, child_type, log):
        if self.uri is None:
            raise core.QTIUnimplementedError("inclusion of inline images")
        if child_type is html.BlockMixin:
            # We must wrap this img in a <p>
            p = parent.add_child(html.P, (qtiv2.core.IMSQTI_NAMESPACE, 'p'))
            img = p.add_child(
                html.Img, (qtiv2.core.IMSQTI_NAMESPACE, 'img'))
        else:
            img = parent.add_child(
                html.Img, (qtiv2.core.IMSQTI_NAMESPACE, 'img'))
        img.src = self.resolve_uri(self.uri)
        if self.width is not None:
            img.width = html.LengthType(self.width)
        if self.height is not None:
            img.height = html.LengthType(self.height)


class MatAudio(core.QTIElement, MatThingMixin):

    """The <mataudio> element is used to contain audio content that is to be
    displayed to the users::

            <!ELEMENT mataudio (#PCDATA)>
            <!ATTLIST mataudio
                    audiotype	CDATA  'audio/base'
                    label		CDATA  #IMPLIED
                    uri			CDATA  #IMPLIED
                    embedded	CDATA  'base64'
                    entityref	ENTITY  #IMPLIED >"""
    XMLNAME = "mataudio"
    XMLCONTENT = xml.XMLMixedContent

    XMLATTR_audiotype = 'audioType'
    XMLATTR_label = 'label'
    XMLATTR_uri = ('uri', html.uri.URI.from_octets, html.to_text)
    XMLATTR_embedded = 'embedded'
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        MatThingMixin.__init__(self)
        self.audioType = 'audio/base'
        self.label = None
        self.uri = None
        self.embedded = 'base64'

    def is_inline(self):
        return True

    def extract_text(self):
        """We cannot extract text from mataudio so we return a simple
        string."""
        return "[sound]"

    def migrate_content_to_v2(self, parent, child_type, log):
        if self.uri is None:
            raise core.QTIUnimplementedError("inclusion of inline audio")
        if child_type is html.BlockMixin:
            # We must wrap this object in a <p>
            p = parent.add_child(html.P, (qtiv2.core.IMSQTI_NAMESPACE, 'p'))
            obj = p.add_child(
                html.Object, (qtiv2.core.IMSQTI_NAMESPACE, 'object'))
        else:
            obj = parent.add_child(
                html.Object, (qtiv2.core.IMSQTI_NAMESPACE, 'object'))
        obj.data = self.resolve_uri(self.uri)
        obj.type = self.audioType


class MatVideo(core.QTIElement, PositionMixin, MatThingMixin):

    """The <matvideo> element is used to contain video content that is to be
    displayed to the users::

            <!ELEMENT matvideo (#PCDATA)>
            <!ATTLIST matvideo
                    videotype	CDATA  'video/avi'
                    label		CDATA  #IMPLIED
                    uri			CDATA  #IMPLIED
                    width		CDATA  #IMPLIED
                    height		CDATA  #IMPLIED
                    y0			CDATA  #IMPLIED
                    x0			CDATA  #IMPLIED
                    embedded	CDATA  'base64'
                    entityref	ENTITY  #IMPLIED >"""
    XMLNAME = "matvideo"
    XMLCONTENT = xml.XMLMixedContent
    XMLATTR_videotype = 'videoType'
    XMLATTR_label = 'label'
    XMLATTR_uri = ('uri', html.uri.URI.from_octets, html.to_text)
    XMLATTR_embedded = 'embedded'
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        MatThingMixin.__init__(self)
        PositionMixin.__init__(self)
        self.videoType = 'video/avi'
        self.label = None
        self.uri = None
        self.embedded = 'base64'

    def is_inline(self):
        return True

    def extract_text(self):
        """We cannot extract text from matvideo so we return a simple
        string."""
        return "[video]"

    def migrate_content_to_v2(self, parent, child_type, log):
        if self.uri is None:
            raise core.QTIUnimplementedError("inclusion of inline video")
        if child_type is html.BlockMixin:
            # We must wrap this object in a <p>
            p = parent.add_child(html.P, (qtiv2.core.IMSQTI_NAMESPACE, 'p'))
            obj = p.add_child(
                html.Object, (qtiv2.core.IMSQTI_NAMESPACE, 'object'))
        else:
            obj = parent.add_child(
                html.Object, (qtiv2.core.IMSQTI_NAMESPACE, 'object'))
        obj.data = self.resolve_uri(self.uri)
        obj.type = self.videoType


class MatApplet(core.QTIElement, PositionMixin, MatThingMixin):

    """The <matapplet> element is used to contain applet content that is to be
    displayed to the users. Parameters that are to be passed to the applet
    being launched should be enclosed in a CDATA block within the content of
    the <matapplet> element::

            <!ELEMENT matapplet (#PCDATA)>
            <!ATTLIST matapplet
                    label		CDATA  #IMPLIED
                    uri			CDATA  #IMPLIED
                    y0			CDATA  #IMPLIED
                    height		CDATA  #IMPLIED
                    width		CDATA  #IMPLIED
                    x0			CDATA  #IMPLIED
                    embedded	CDATA  'base64'
                    entityref	ENTITY  #IMPLIED >"""
    XMLNAME = "matapplet"
    XMLCONTENT = xml.XMLMixedContent
    XMLATTR_label = 'label'
    XMLATTR_uri = ('uri', html.uri.URI.from_octets, html.to_text)
    XMLATTR_embedded = 'embedded'
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        MatThingMixin.__init__(self)
        PositionMixin.__init__(self)
        self.label = None
        self.uri = None
        self.embedded = 'base64'

    def is_inline(self):
        return True

    def extract_text(self):
        """We cannot extract text from matapplet so we return a simple
        string."""
        return "[applet]"

    def migrate_content_to_v2(self, parent, child_type, log):
        raise core.QTIUnimplementedError("matapplet")


class MatApplication(core.QTIElement, MatThingMixin):

    """The <matapplication> element is used to contain application content that
    is to be displayed to the users. Parameters that are to be passed to the
    application being launched should be enclosed in a CDATA block within the
    content of the <matapplication> element::

            <!ELEMENT matapplication (#PCDATA)>
            <!ATTLIST matapplication
                    apptype		CDATA  #IMPLIED
                    label		CDATA  #IMPLIED
                    uri			CDATA  #IMPLIED
                    embedded	CDATA  'base64'
                    entityref	ENTITY  #IMPLIED >"""
    XMLNAME = "matapplication"
    XMLCONTENT = xml.XMLMixedContent
    XMLATTR_apptype = 'appType'
    XMLATTR_label = 'label'
    XMLATTR_uri = ('uri', html.uri.URI.from_octets, html.to_text)
    XMLATTR_embedded = 'embedded'
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        MatThingMixin.__init__(self)
        self.appType = None
        self.label = None
        self.uri = None
        self.embedded = 'base64'

    def is_inline(self):
        return True

    def extract_text(self):
        """We cannot extract text from matapplication so we return a simple
        string."""
        return "[application]"

    def migrate_content_to_v2(self, parent, child_type, log):
        raise core.QTIUnimplementedError("matapplication")


class MatRef(MatThingMixin, core.QTIElement):

    """The <matref> element is used to contain a reference to the required
    material. This material will have had an identifier assigned to enable such
    a reference to be reconciled when the instance is parsed into the system.
    <matref> should only be used to reference a material component and not a
    <material> element (the element <material_ref> should be used for the
    latter)::

            <!ELEMENT matref EMPTY>
            <!ATTLIST matref linkrefid CDATA  #REQUIRED >"""
    XMLNAME = "matref"
    XMLATTR_linkrefid = 'linkRefID'
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        MatThingMixin.__init__(self)
        self.linkRefID = None

    def find_mat_thing(self):
        mat_thing = None
        doc = self.get_document()
        if doc:
            mat_thing = doc.find_mat_thing(self.linkRefID)
        if mat_thing is None:
            raise core.QTIError("Bad matref: %s" % str(self.linkRefID))
        return mat_thing

    def is_inline(self):
        return self.find_mat_thing().is_inline()

    def extract_text(self):
        return self.find_mat_thing().extract_text()

    def migrate_content_to_v2(self, parent, child_type, log):
        self.find_mat_thing().migrate_content_to_v2(parent, child_type, log)


class MatExtension(core.QTIElement, MatThingMixin):

    """The extension facility to enable proprietary types of material to be
    included with the corresponding data object::

            <!ELEMENT mat_extension ANY>"""
    XMLNAME = "mat_extension"
    XMLCONTENT = xml.XMLMixedContent


class FlowMixin(object):

    """Mix-in class to identify all flow elements::

            ( flow | flow_mat | flow_label)"""
    pass


class FlowMatContainer(ContentMixin, QTICommentContainer):

    """Abstract class used to represent objects that contain flow_mat::

            <!ELEMENT XXXXXXXXXX (qticomment? , (material+ | flow_mat+))>"""

    def __init__(self, parent):
        QTICommentContainer.__init__(self, parent)
        ContentMixin.__init__(self)

    def content_child(self, child_class):
        return issubclass(child_class, (Material, FlowMat))

    def get_children(self):
        return itertools.chain(
            QTICommentContainer.get_children(self),
            ContentMixin.get_content_children(self))


class FlowMat(FlowMatContainer, FlowMixin):

    """This element allows the materials to be displayed to the users to be
    grouped together using flows. The manner in which these flows are handled
    is dependent upon the display-engine::

            <!ELEMENT flow_mat (qticomment? , (flow_mat | material |
            material_ref)+)>
            <!ATTLIST flow_mat  class CDATA  'Block' >"""
    XMLNAME = "flow_mat"
    XMLATTR_class = ('flow_class', None, None, list)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        FlowMatContainer.__init__(self, parent)

    def content_child(self, child_class):
        return issubclass(child_class, MaterialRef) or \
            FlowMatContainer.content_child(self, child_class)

    def is_inline(self):
        """flowmat is always treated as a block if flow_class is specified, otherwise
        it is treated as a block unless it is an only child."""
        if not self.flow_class:
            return self.inline_children()
        else:
            return False

    def migrate_content_to_v2(self, parent, child_type, log):
        """flow typically maps to a div element.

        A flow with a specified class always becomes a div."""
        if self.flow_class:
            if child_type in (html.BlockMixin, html.FlowMixin):
                div = parent.add_child(
                    html.Div, (qtiv2.core.IMSQTI_NAMESPACE, 'div'))
                div.style_class = self.flow_class
                FlowMatContainer.migrate_content_to_v2(
                    self, div, html.FlowMixin, log)
            else:
                span = parent.add_child(
                    html.Span, (qtiv2.core.IMSQTI_NAMESPACE, 'span'))
                span.style_class = self.flow_class
                FlowMatContainer.migrate_content_to_v2(
                    self, span, html.InlineMixin, log)
        else:
            # ignore the flow, br handling is done automatically by the parent
            FlowMatContainer.migrate_content_to_v2(self, parent, child_type,
                                                   log)


class MetadataContainerMixin(object):

    """A mix-in class used to hold dictionaries of metadata.

    There is a single dictionary maintained to hold all metadata values, each
    value is a list of tuples of the form (value string, defining element).
    Values are keyed on the field label or tag name with any leading qmd\_
    prefix removed."""

    def __init__(self):
        self.metadata = {}

    def declare_metadata(self, label, entry, definition=None):
        label = label.lower()
        if label[:4] == "qmd_":
            label = label[4:]
        if label not in self.metadata:
            self.metadata[label] = []
        self.metadata[label].append((entry, definition))


class QTIMetadata(core.QTIElement):

    """The container for all of the vocabulary-based QTI-specific meta-data.
    This structure is available to each of the four core ASI data structures::

    <!ELEMENT qtimetadata (vocabulary? , qtimetadatafield+)>"""
    XMLNAME = 'qtimetadata'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.Vocabulary = None
        self.QTIMetadataField = []

    def get_children(self):
        if self.Vocabulary:
            yield self.Vocabulary
        for child in self.QTIMetadataField:
            yield child


class Vocabulary(core.QTIElement):

    """The vocabulary to be applied to the associated meta-data fields. The
    vocabulary is defined either using an external file or it is included as a
    comma separated list::

            <!ELEMENT vocabulary (#PCDATA)>
            <!ATTLIST vocabulary
                    uri			CDATA  #IMPLIED
                    entityref	ENTITY  #IMPLIED
                    vocab_type	CDATA  #IMPLIED >"""
    XMLNAME = "vocabulary"
    XMLATTR_entityref = 'entityRef'
    XMLATTR_uri = 'uri'
    XMLATTR_vocab_type = 'vocabType'

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.uri = None
        self.entityRef = None
        self.vocabType = None


class QTIMetadataField(core.QTIElement):

    """The structure responsible for containing each of the QTI-specific
    meta-data fields::

            <!ELEMENT qtimetadatafield	(fieldlabel , fieldentry)>
            <!ATTLIST qtimetadatafield  xml:lang CDATA  #IMPLIED >"""
    XMLNAME = 'qtimetadatafield'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.FieldLabel = FieldLabel(self)
        self.FieldEntry = FieldEntry(self)

    def get_children(self):
        yield self.FieldLabel
        yield self.FieldEntry

    def content_changed(self):
        label = self.FieldLabel.get_value()
        label = {'marks': 'maximumscore',
                 'qmd_marks': 'maximumscore',
                 'name': 'title',
                 'qmd_name': 'title',
                 'syllabusarea': 'topic',
                 'qmd_syllabusarea': 'topic',
                 'item type': 'itemtype',
                 'question type': 'itemtype',
                 'qmd_layoutstatus': 'status',
                 'layoutstatus': 'status'}.get(label, label)
        # Still need to handle creator and owner
        self.declare_metadata(label, self.FieldEntry.get_value(), self)


class FieldLabel(core.QTIElement):

    """Used to contain the name of the QTI-specific meta-data field::

            <!ELEMENT fieldlabel (#PCDATA)>"""
    XMLNAME = "fieldlabel"
    XMLCONTENT = xml.XMLMixedContent


class FieldEntry(core.QTIElement):

    """Used to contain the actual data entry of the QTI-specific meta-data field
    named using the associated 'fieldlabel' element::

            <!ELEMENT fieldentry (#PCDATA)>"""
    XMLNAME = "fieldentry"
    XMLCONTENT = xml.XMLMixedContent


class Objectives(FlowMatContainer):

    """The objectives element is used to store the information that describes
    the educational aims of the Item. These objectives can be defined for each
    of the different 'view' perspectives. This element should not be used to
    contain information specific to an Item because the question-engine may not
    make this information available to the Item during the actual test::

            <!ELEMENT objectives (qticomment? , (material+ | flow_mat+))>
            <!ATTLIST objectives  view	(All | Administrator | AdminAuthority |
                                    Assessor | Author | Candidate |
                                    InvigilatorProctor | Psychometrician |
                                    Scorer | Tutor ) 'All' >"""
    XMLNAME = 'objectives'
    XMLATTR_view = ('view', core.View.from_str_lower, core.View.to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        FlowMatContainer.__init__(self, parent)
        self.view = core.View.DEFAULT

    def migrate_to_v2(self, v2_item, log):
        """Adds rubric representing these objectives to the given item's
        body"""
        rubric = v2_item.add_child(
            qtiv2.content.ItemBody).add_child(qtiv2.content.RubricBlock)
        rubric.set_attribute(
            'view',
            qtiv2.core.View.list_to_str(
                core.MigrateV2View(
                    self.view,
                    log)))
        # rubric is not a flow-container so we force inlines to be p-wrapped
        self.migrate_content_to_v2(rubric, html.BlockMixin, log)

    def migrate_objectives_to_lrm(self, lom, log):
        """Adds educational description from these objectives."""
        description, lang = self.extract_text()
        edu_description = lom.add_child(
            imsmd.LOMEducational).add_child(imsmd.Description)
        edu_description.AddString(lang, description)


class Rubric(FlowMatContainer):

    """The rubric element is used to contain contextual information that is
    important to the element e.g. it could contain standard data values that
    might or might not be useful for answering the question. Different sets of
    rubric can be defined for each of the possible 'views'. The material
    contained within the rubric must be displayed to the participant::

            <!ELEMENT rubric (qticomment? , (material+ | flow_mat+))>
            <!ATTLIST rubric  view	(All | Administrator | AdminAuthority |
                                    Assessor | Author | Candidate |
                                    InvigilatorProctor | Psychometrician |
                                    Scorer | Tutor ) 'All' >"""
    XMLNAME = 'rubric'
    XMLATTR_view = ('view', core.View.from_str_lower, core.View.to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        FlowMatContainer.__init__(self, parent)
        self.view = core.View.DEFAULT

    def migrate_to_v2(self, v2_item, log):
        if self.view == core.View.All:
            log.append(
                'Warning: rubric with view="All" replaced by <div> with'
                ' class="rubric"')
            rubric = v2_item.get_or_add_child(
                qtiv2.content.ItemBody).add_child(
                    html.Div, (qtiv2.core.IMSQTI_NAMESPACE, 'div'))
            rubric.style_class = ['rubric']
        else:
            rubric = v2_item.get_or_add_child(
                qtiv2.content.ItemBody).add_child(qtiv2.content.RubricBlock)
            rubric.set_attribute(
                'view',
                qtiv2.core.View.list_to_str(
                    core.MigrateV2View(
                        self.view,
                        log)))
        # rubric is not a flow-container so we force inlines to be p-wrapped
        self.migrate_content_to_v2(rubric, html.BlockMixin, log)


class DecVar(core.QTIElement):

    """The <decvar> element permits the declaration of the scoring variables
    ::

            <!ELEMENT decvar (#PCDATA)>
            <!ATTLIST decvar  varname CDATA  'SCORE' ::
                    vartype		(Integer |  String |  Decimal |  Scientific |
                            Boolean | Enumerated | Set )  'Integer'
                    defaultval 	CDATA  #IMPLIED
                    minvalue   	CDATA  #IMPLIED
                    maxvalue   	CDATA  #IMPLIED
                    members    	CDATA  #IMPLIED
                    cutvalue	CDATA  #IMPLIED >"""
    XMLNAME = "decvar"
    XMLATTR_cutvalue = 'cutValue'
    XMLATTR_defaultval = 'defaultValue'
    XMLATTR_maxvalue = 'maxValue'
    XMLATTR_members = 'members'
    XMLATTR_minvalue = 'minValue'
    XMLATTR_varname = 'varName'
    XMLATTR_vartype = (
        'varType', core.VarType.from_str_title, core.VarType.to_str)
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.varName = 'SCORE'
        self.varType = core.VarType.Integer
        self.defaultValue = None
        self.minValue = None
        self.maxValue = None
        self.members = None
        self.cutValue = None

    def migrate_to_v2(self, v2_item, log):
        d = v2_item.add_child(qtiv2.variables.OutcomeDeclaration)
        v2type = core.MigrateV2VarType(self.varType, log)
        if self.varType == core.VarType.Set:
            log.append(
                'Warning: treating vartype="Set" as equivalent to'
                ' "Enumerated"')
        elif v2type is None:
            log.append(
                'Error: bad vartype for decvar "%s"; defaulting to integer' %
                self.varName)
            v2type = qtiv2.variables.BaseType.integer
        d.baseType = v2type
        d.cardinality = qtiv2.variables.Cardinality.single
        d.identifier = qtiv2.core.ValidateIdentifier(self.varName)
        if self.defaultValue is not None:
            value = d.add_child(qtiv2.variables.DefaultValue).add_child(
                qtiv2.variables.ValueElement)
            value.set_value(self.defaultValue)
        # to do... min and max were actually constraints in QTI v1 ...
        # so we will need to fix these up later with responseConditions to
        # constraint the value
        if self.minValue is not None:
            d.normalMinimum = float(self.minValue)
        if self.maxValue is not None:
            d.normalMaximum = float(self.maxValue)
        if self.members is not None:
            log.append(
                'Warning: enumerated members no longer supported,'
                ' ignoring "%s"' %
                self.members)
        if v2type in (qtiv2.variables.BaseType.integer,
                      qtiv2.variables.BaseType.float):
            # we need to adjust minValue/maxValue later
            if self.cutValue is not None:
                d.masteryValue = float(self.cutValue)
        v2_item.RegisterDeclaration(d)

    def content_changed(self):
        """The decvar element is supposed to be empty but QTI v1 content is all
        over the place."""
        try:
            value = self.get_value()
            if value is not None:
                assert value.strip() == ''
        except xml.XMLMixedContentError:
            # Even more confusing
            pass


class InterpretVar(ContentMixin, core.QTIElement):

    """The <interpretvar> element is used to provide statistical interpretation
    information about the associated variables::

            <!ELEMENT interpretvar (material | material_ref)>
            <!ATTLIST interpretvar
                    view	(All | Administrator | AdminAuthority | Assessor |
                            Author | Candidate | InvigilatorProctor |
                            Psychometrician | Scorer | Tutor )  'All'
                            varname CDATA  'SCORE' >"""
    XMLNAME = "interpretvar"
    XMLATTR_view = ('view', core.View.from_str_lower, core.View.to_str)
    XMLATTR_varname = 'varName'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        ContentMixin.__init__(self)
        self.view = core.View.DEFAULT
        self.varName = 'SCORE'

    def content_child(self, child_class):
        return issubclass(child_class, (Material, MaterialRef))

    def get_children(self):
        return ContentMixin.get_content_children(self)

    def migrate_to_v2(self, v2_item, log):
        identifier = qtiv2.core.ValidateIdentifier(self.varName)
        if self.view != core.View.All:
            log.append(
                'Warning: view restriction on outcome interpretation no longer'
                ' supported (%s)' %
                core.View.to_str(
                    self.view))
        d = v2_item.declarations.get(identifier)
        di, lang = self.extract_text()
        di = xsi.white_space_collapse(di)
        if d.interpretation:
            d.interpretation = d.interpretation + "; " + di
        else:
            d.interpretation = di
        # we drop the lang as this isn't supported on declarations


class SetVar(core.QTIElement):

    """The <setvar> element is responsible for changing the value of the scoring
    variable as a result of the associated response processing test::

            <!ELEMENT setvar (#PCDATA)>
            <!ATTLIST setvar  varname CDATA  'SCORE'
                    action     (Set | Add | Subtract | Multiply | Divide )
                    'Set' >"""
    XMLNAME = "setvar"
    XMLATTR_varname = 'varName'
    XMLATTR_action = (
        'action', core.Action.from_str_title, core.Action.to_str)
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.varName = 'SCORE'
        self.action = core.Action.DEFAULT

    def migrate_rule_to_v2(self, parent, log):
        v2_item = parent.find_parent(qtiv2.items.AssessmentItem)
        identifier = qtiv2.core.ValidateIdentifier(self.varName)
        outcome = v2_item.declarations.get(identifier, None)
        if outcome is None:
            raise core.QTIUnimplementedError("Auto-declared outcomes")
        set_value = parent.add_child(qtiv2.processing.SetOutcomeValue)
        set_value.identifier = identifier
        if outcome.cardinality != qtiv2.variables.Cardinality.single:
            raise core.QTIUnimplementedError(
                "setvar for '%s' with cardinality %s" %
                (identifier,
                 qtiv2.variables.Cardinality.encode[
                     outcome.cardinality]))
        value = None
        variable = None
        if not self.action or self.action == core.Action.Set:
            value = set_value.add_child(qtiv2.expressions.BaseValue)
        else:
            if self.action == core.Action.Add:
                op = set_value.add_child(qtiv2.expressions.Sum)
            elif self.action == core.Action.Subtract:
                op = set_value.add_child(qtiv2.expressions.Subtract)
            elif self.action == core.Action.Multiply:
                op = set_value.add_child(qtiv2.expressions.Product)
            elif self.action == core.Action.Divide:
                op = set_value.add_child(qtiv2.expressions.Divide)
            variable = op.add_child(qtiv2.expressions.Variable)
            variable.identifier = identifier
            value = op.add_child(qtiv2.expressions.BaseValue)
        value.baseType = outcome.baseType
        value.set_value(self.get_value().strip())


class DisplayFeedback(core.QTIElement):

    """The <displayfeedback> element is responsible for assigning an associated
    feedback to the response processing if the 'True' state is created through
    the associated response processing condition test::

            <!ELEMENT displayfeedback (#PCDATA)>
            <!ATTLIST displayfeedback
                    feedbacktype	(Response | Solution | Hint )  'Response'
                    linkrefid		CDATA  #REQUIRED >"""
    XMLNAME = "displayfeedback"
    XMLATTR_feedbacktype = (
        'feedbackType',
        core.FeedbackType.from_str_title,
        core.FeedbackType.to_str)
    XMLATTR_linkrefid = 'linkRefID'
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.feedbackType = core.FeedbackType.DEFAULT
        self.linkRefID = None

    def migrate_rule_to_v2(self, parent, log):
        v2_item = parent.find_parent(qtiv2.items.AssessmentItem)
        qtiv2.core.ValidateIdentifier(self.linkRefID, 'FEEDBACK_')
        outcome = v2_item.declarations.get('FEEDBACK', None)
        if outcome is None:
            d = v2_item.add_child(qtiv2.variables.OutcomeDeclaration)
            d.baseType = qtiv2.variables.BaseType.identifier
            d.cardinality = qtiv2.variables.Cardinality.multiple
            d.identifier = 'FEEDBACK'
            v2_item.RegisterDeclaration(d)
        set_value = parent.add_child(qtiv2.processing.SetOutcomeValue)
        set_value.identifier = 'FEEDBACK'
        multiple = set_value.add_child(qtiv2.expressions.Multiple)
        variable = multiple.add_child(qtiv2.expressions.Variable)
        variable.identifier = 'FEEDBACK'
        value = multiple.add_child(qtiv2.expressions.BaseValue)
        value.baseType = qtiv2.variables.BaseType.identifier
        value.set_value(self.linkRefID)


class ConditionVar(core.QTIElement):

    """The conditional test that is to be applied to the user's response. A wide
    range of separate and combinatorial test can be applied::

            <!ELEMENT conditionvar (not | and | or | unanswered | other |
                    varequal | varlt | varlte | vargt | vargte | varsubset |
                    varinside | varsubstring | durequal | durlt | durlte |
                    durgt | durgte | var_extension)+>"""
    XMLNAME = "conditionvar"
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.ExtendableExpressionMixin = []

    def get_children(self):
        return iter(self.ExtendableExpressionMixin)

    def migrate_expression_to_v2(self, parent, log):
        if len(self.ExtendableExpressionMixin) > 1:
            # implicit and
            eand = parent.add_child(qtiv2.expressions.And)
            for ie in self.ExtendableExpressionMixin:
                ie.migrate_expression_to_v2(eand, log)
        elif self.ExtendableExpressionMixin:
            self.ExtendableExpressionMixin[0].migrate_expression_to_v2(
                parent, log)
        else:
            log.append("Warning: empty condition replaced with null operator")
            parent.add_child(qtiv2.expressions.Null)


class ExtendableExpressionMixin:

    """Abstract mixin class to indicate an expression, including
    var_extension"""

    def migrate_expression_to_v2(self, parent, log):
        raise core.QTIUnimplementedError(
            "Expression element: %s" % self.__class__.__name__)


class ExpressionMixin(ExtendableExpressionMixin):

    """Abstract mixin class to indicate an expression excluding
    var_extension"""
    pass


class VarThing(core.QTIElement, ExpressionMixin):

    """Abstract class for var\\* elements
    ::

            <!ATTLIST *
                    respident	CDATA #REQUIRED
                    index		CDATA  #IMPLIED >"""
    XMLATTR_respident = 'responseIdentifier'
    XMLATTR_index = ('index', xsi.integer_from_str, xsi.integer_to_str)
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.responseIdentifier = ''
        self.index = None

    def migrate_missing_to_v2(self, identifier, parent, log):
        log.append(
            "Warning: test of undeclared response (%s)"
            " replaced with Null operator" %
            identifier)
        parent.add_child(qtiv2.expressions.Null)

    def migrate_variable_to_v2(self, d, parent, log):
        if self.index:
            if d.cardinality == qtiv2.variables.Cardinality.multiple:
                log.append(
                    "Warning: index ignored for response variable of"
                    " cardinality multiple")
            elif d.cardinality == qtiv2.variables.Cardinality.single:
                log.append(
                    "Warning: index ignored for response variable of"
                    " cardinality single")
            else:
                parent = parent.add_child(qtiv2.expressions.Index)
                parent.n = self.index
        var_expression = parent.add_child(qtiv2.expressions.Variable)
        var_expression.identifier = d.identifier

    def migrate_value_to_v2(self, d, parent, log):
        value = self.get_value().strip()
        if d.baseType in (qtiv2.variables.BaseType.pair,
                          qtiv2.variables.BaseType.directedPair,
                          qtiv2.variables.BaseType.point):
            value = value.replace(',', ' ')
        elif d.baseType == qtiv2.variables.BaseType.identifier:
            value = qtiv2.core.ValidateIdentifier(value)
        bv = parent.add_child(qtiv2.expressions.BaseValue)
        bv.baseType = d.baseType
        bv.set_value(value)


class VarEqual(VarThing):

    """The <varequal> element is the test of equivalence. The data for the test
    is contained within the element's PCDATA string and must be the same as one
    of the <response_label> values (this were assigned using the ident
    attribute)::

            <!ELEMENT varequal (#PCDATA)>
            <!ATTLIST varequal
                    case  (Yes | No )  'No'
                    respident CDATA  #REQUIRED"
                    index CDATA  #IMPLIED >"""
    XMLNAME = "varequal"
    XMLATTR_case = ('case', core.ParseYesNo, core.FormatYesNo)

    def __init__(self, parent):
        VarThing.__init__(self, parent)
        self.case = False

    def migrate_expression_to_v2(self, parent, log):
        v2_item = parent.find_parent(qtiv2.items.AssessmentItem)
        identifier = qtiv2.core.ValidateIdentifier(self.responseIdentifier)
        d = v2_item.declarations.get(identifier, None)
        if d is None:
            self.migrate_missing_to_v2(identifier, parent, log)
        elif d.cardinality == qtiv2.variables.Cardinality.single:
            # simple test of equality
            if (d.baseType == qtiv2.variables.BaseType.identifier or
                    d.baseType == qtiv2.variables.BaseType.pair):
                if not self.case:
                    log.append(
                        "Warning: case-insensitive comparison of identifiers"
                        " not supported in version 2")
                expression = parent.add_child(qtiv2.expressions.Match)
            elif d.baseType == qtiv2.variables.BaseType.integer:
                expression = parent.add_child(qtiv2.expressions.Match)
            elif d.baseType == qtiv2.variables.BaseType.string:
                expression = parent.add_child(qtiv2.expressions.StringMatch)
                expression.caseSensitive = self.case
            elif d.baseType == qtiv2.variables.BaseType.float:
                log.append(
                    "Warning: equality operator with float values is"
                    " deprecated")
                expression = parent.add_child(qtiv2.expressions.Equal)
            else:
                raise QTIUnimplementedOperator(
                    "varequal(%s)" %
                    qtiv2.variables.BaseType.Encode(
                        d.baseType))
            self.migrate_variable_to_v2(d, expression, log)
            self.migrate_value_to_v2(d, expression, log)
        else:
            # This test simply becomes a member-test operation
            if (d.baseType == qtiv2.variables.BaseType.identifier or
                    qtiv2.variables.BaseType.pair):
                if not self.case:
                    log.append(
                        "Warning: case-insensitive comparison of identifiers"
                        " not supported in version 2")
            elif d.baseType == qtiv2.variables.BaseType.string:
                if not self.case:
                    log.append(
                        "Warning: member operation cannot be case-insensitive"
                        " when baseType is string")
            elif d.baseType == qtiv2.variables.BaseType.float:
                log.append(
                    "Warning: member operation is deprecated when"
                    " baseType is float")
            else:
                raise QTIUnimplementedOperator(
                    "varequal(%s)" %
                    qtiv2.variables.BaseType.Encode(
                        d.baseType))
            expression = parent.add_child(qtiv2.expressions.Member)
            self.migrate_value_to_v2(d, expression, log)
            self.migrate_variable_to_v2(d, expression, log)


class VarInequality(VarThing):

    """Abstract class for varlt, varlte, vargt and vargte."""

    def migrate_inequality_to_v2(self):
        """Returns the class to use in qtiv2"""
        raise QTIUnimplementedOperator(self.xmlname)

    def migrate_expression_to_v2(self, parent, log):
        v2_item = parent.find_parent(qtiv2.items.AssessmentItem)
        identifier = qtiv2.core.ValidateIdentifier(self.responseIdentifier)
        d = v2_item.declarations.get(identifier, None)
        if d is None:
            self.migrate_missing_to_v2(identifier, parent, log)
        elif d.cardinality == qtiv2.variables.Cardinality.single:
            # simple inequality
            if (d.baseType == qtiv2.variables.BaseType.integer or
                    d.baseType == qtiv2.variables.BaseType.float):
                expression = parent.add_child(self.migrate_inequality_to_v2())
            else:
                raise QTIUnimplementedOperator(
                    "%s(%s)" %
                    (self.xmlname,
                     qtiv2.variables.BaseType.Encode(
                         d.baseType)))
            self.migrate_variable_to_v2(d, expression, log)
            self.migrate_value_to_v2(d, expression, log)
        else:
            raise QTIUnimplementedOperator(
                "%s(%s:%s)" %
                (self.xmlname, qtiv2.variables.Cardinality.Encode(
                    d.cardinality), qtiv2.variables.BaseType.Encode(
                    d.baseType)))


class VarLT(VarInequality):

    """The <varlt> element is the 'less than' test. The data for the test is
    contained within the element's PCDATA string and is assumed to be numerical
    in nature::

            <!ELEMENT varlt (#PCDATA)>
            <!ATTLIST varlt
                    respident	CDATA  #REQUIRED"
                    index		CDATA  #IMPLIED >"""
    XMLNAME = "varlt"
    XMLCONTENT = xml.XMLMixedContent

    def migrate_inequality_to_v2(self):
        return qtiv2.expressions.LT


class VarLTE(VarInequality):

    """The <varlte> element is the 'less than or equal' test. The data for the
    test is contained within the element's PCDATA string and is assumed to be
    numerical in nature::

            <!ELEMENT varlte (#PCDATA)>
            <!ATTLIST varlte
                    respident CDATA  #REQUIRED"
                    index CDATA  #IMPLIED >"""
    XMLNAME = "varlte"
    XMLCONTENT = xml.XMLMixedContent

    def migrate_inequality_to_v2(self):
        return qtiv2.expressions.LTE


class VarGT(VarInequality):

    """The <vargt> element is the 'greater than' test. The data for the test is
    contained within the element's PCDATA string and is assumed to be numerical
    in nature::

            <!ELEMENT vargt (#PCDATA)>
            <!ATTLIST vargt
                    respident CDATA  #REQUIRED"
                    index CDATA  #IMPLIED >"""
    XMLNAME = "vargt"
    XMLCONTENT = xml.XMLMixedContent

    def migrate_inequality_to_v2(self):
        return qtiv2.expressions.GT


class VarGTE(VarInequality):

    """The <vargte> element is the 'greater than or equal to' test. The data for
    the test is contained within the element's PCDATA string and is assumed to
    be numerical in nature::

            <!ELEMENT vargte (#PCDATA)>
            <!ATTLIST vargte
                    respident CDATA  #REQUIRED"
                    index CDATA  #IMPLIED >"""
    XMLNAME = "vargte"
    XMLCONTENT = xml.XMLMixedContent

    def migrate_inequality_to_v2(self):
        return qtiv2.expressions.GTE


class VarSubset(core.QTIElement, ExpressionMixin):

    """The <varsubset> element is the 'member of a list/set' test. The data for
    the test is contained within the element's PCDATA string. The set is a
    comma separated list with no enclosing parentheses::

            <!ELEMENT varsubset (#PCDATA)>
            <!ATTLIST varsubset
                    respident CDATA  #REQUIRED"
                    setmatch     (Exact | Partial )  'Exact'
                    index CDATA  #IMPLIED >"""
    XMLNAME = "varsubset"
    XMLCONTENT = xml.XMLMixedContent


class VarSubString(core.QTIElement, ExpressionMixin):

    """The <varsubstring> element is used to determine if a given string is a
    substring of some other string::

            <!ELEMENT varsubstring (#PCDATA)>
            <!ATTLIST varsubstring
                    index CDATA  #IMPLIED
                    respident CDATA  #REQUIRED"
                    case  (Yes | No )  'No' >"""
    XMLNAME = "varsubstring"
    XMLCONTENT = xml.XMLMixedContent


class VarInside(VarThing):

    """The <varinside> element is the 'xy-co-ordinate inside an area' test. The
    data for the test is contained within the element's PCDATA string and is a
    set of co-ordinates that define the area::

            <!ELEMENT varinside (#PCDATA)>
            <!ATTLIST varinside
                    areatype     (Ellipse | Rectangle | Bounded )  #REQUIRED
                    respident CDATA  #REQUIRED"
                    index CDATA  #IMPLIED >"""
    XMLNAME = "varinside"
    XMLATTR_areatype = (
        'areaType', core.Area.from_str_title, core.Area.to_str)
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        VarThing.__init__(self, parent)
        self.areaType = None

    def migrate_expression_to_v2(self, parent, log):
        v2_item = parent.find_parent(qtiv2.items.AssessmentItem)
        identifier = qtiv2.core.ValidateIdentifier(self.responseIdentifier)
        d = v2_item.declarations.get(identifier, None)
        if d is None:
            self.migrate_missing_to_v2(identifier, parent, log)
        elif d.cardinality == qtiv2.variables.Cardinality.single:
            # is the point in the area?
            if d.baseType == qtiv2.variables.BaseType.point:
                expression = parent.add_child(qtiv2.expressions.Inside)
                expression.shape, expression.coords = core.MigrateV2AreaCoords(
                    self.areaType, self.get_value(), log)
                self.migrate_variable_to_v2(d, expression, log)
            else:
                raise core.QTIUnimplementedError(
                    "varinside(%s)" %
                    qtiv2.variables.BaseType.to_str(
                        d.baseType))
        else:
            raise core.QTIUnimplementedError(
                "varinside with multiple/orderd variable")


class DurEqual(core.QTIElement, ExpressionMixin):

    """The <durequal> element is the 'duration equal to' test i.e. a test on the
    time taken to make the response::

            <!ELEMENT durequal (#PCDATA)>
            <!ATTLIST durequal
                    index CDATA  #IMPLIED
                    respident CDATA  #REQUIRED" >"""
    XMLNAME = "durequal"
    XMLCONTENT = xml.XMLMixedContent


class DurLT(core.QTIElement, ExpressionMixin):

    """The <durlt> element is the 'duration less than' test i.e. a test on the
    time taken to make the response::

            <!ELEMENT durlt (#PCDATA)>
            <!ATTLIST durlt
                    index		CDATA  #IMPLIED
                    respident	CDATA  #REQUIRED" >"""
    XMLNAME = "durlt"
    XMLCONTENT = xml.XMLMixedContent


class DurLTE(core.QTIElement, ExpressionMixin):

    """The <durlte> element is the 'duration less than or equal to' test i.e. a
    test on the time taken to make the response::

            <!ELEMENT durlte (#PCDATA)>
            <!ATTLIST durlte
                    index		CDATA  #IMPLIED
                    respident	CDATA  #REQUIRED" >"""
    XMLNAME = "durlte"
    XMLCONTENT = xml.XMLMixedContent


class DurGT(core.QTIElement, ExpressionMixin):

    """The <durgt> element is the 'duration greater than' test i.e. a test on
    the time taken to make the response::

            <!ELEMENT durgt (#PCDATA)>
            <!ATTLIST durgt
                    index		CDATA  #IMPLIED
                    respident	CDATA  #REQUIRED" >"""
    XMLNAME = "durgt"
    XMLCONTENT = xml.XMLMixedContent


class DurGTE(core.QTIElement, ExpressionMixin):

    """The <durgte> element is the 'duration greater than or equal to' test i.e.
    a test on the time taken to make the response::

            <!ELEMENT durgte (#PCDATA)>
            <!ATTLIST durgte
                    index		CDATA  #IMPLIED
                    respident	CDATA  #REQUIRED" >"""
    XMLNAME = "durgte"
    XMLCONTENT = xml.XMLMixedContent


class Not(core.QTIElement, ExpressionMixin):

    """The <not> element inverts the logical test outcome that is required. In
    the case of the <varequal> element produces a 'not equals' test::

            <!ELEMENT not (and | or | not | unanswered | other | varequal |
                    varlt | varlte | vargt | vargte | varsubset | varinside |
                    varsubstring | durequal | durlt | durlte | durgt |
                    durgte)>"""
    XMLNAME = "not"
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.ExpressionMixin = None

    def get_children(self):
        if self.ExpressionMixin:
            yield self.ExpressionMixin

    def migrate_expression_to_v2(self, parent, log):
        if self.ExpressionMixin is None:
            log.append(
                "Warning: empty not condition replaced with null operator")
            parent.add_child(qtiv2.expressions.Null)
        else:
            enot = parent.add_child(qtiv2.expressions.Not)
            self.ExpressionMixin.migrate_expression_to_v2(enot, log)


class And(core.QTIElement, ExpressionMixin):

    """The <and> element is used to create the Boolean 'AND' operation between
    the two or more enclosed tests. The result 'True' is returned if all of the
    tests return a 'True' value::

            <!ELEMENT and (not | and | or | unanswered | other | varequal |
                    varlt | varlte | vargt | vargte | varsubset | varinside |
                    varsubstring | durequal | durlt | durlte | durgt |
                    durgte)+>"""
    XMLNAME = "and"
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.ExpressionMixin = []

    def get_children(self):
        return iter(self.ExpressionMixin)

    def migrate_expression_to_v2(self, parent, log):
        if len(self.ExpressionMixin):
            eand = parent.add_child(qtiv2.expressions.And)
            for e in self.ExpressionMixin:
                e.migrate_expression_to_v2(eand, log)
        else:
            log.append(
                "Warning: empty and condition replaced with null operator")
            parent.add_child(qtiv2.expressions.Null)


class Or(core.QTIElement, ExpressionMixin):

    """The <or> element is used to create the Boolean 'OR' operation between the
    two or more enclosed tests. The result 'True' is returned if one or more of
    the tests return a 'True' value::

            <!ELEMENT or (not | and | or | unanswered | other | varequal |
                    varlt | varlte | vargt | vargte | varsubset | varinside |
                    varsubstring | durequal | durlt | durlte | durgt |
                    durgte)+>"""
    XMLNAME = "or"
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.ExpressionMixin = []

    def get_children(self):
        return iter(self.ExpressionMixin)

    def migrate_expression_to_v2(self, parent, log):
        if len(self.ExpressionMixin):
            eor = parent.add_child(qtiv2.expressions.Or)
            for e in self.ExpressionMixin:
                e.migrate_expression_to_v2(eor, log)
        else:
            log.append(
                "Warning: empty or condition replaced with null operator")
            parent.add_child(qtiv2.expressions.Null)


class Unanswered(core.QTIElement, ExpressionMixin):

    """The <unanswered> element is the condition to be applied if a response is
    not received for the Item i.e. it is unanswered::

            <!ELEMENT unanswered (#PCDATA)>
            <!ATTLIST unanswered  respident CDATA  #REQUIRED" >"""
    XMLNAME = "unanswered"
    XMLCONTENT = xml.XMLMixedContent


class Other(core.QTIElement, ExpressionMixin):

    """The <other> element is used to trigger the condition when all of the
    other tests have not returned a 'True' state::

            <!ELEMENT other (#PCDATA)>"""
    XMLNAME = "other"
    XMLCONTENT = xml.XMLMixedContent

    def migrate_expression_to_v2(self, parent, log):
        log.append(
            "Warning: replacing <other/> with the base value true"
            " - what did you want me to do??")
        bv = parent.add_child(qtiv2.expressions.BaseValue)
        bv.baseType = qtiv2.variables.BaseType.boolean
        bv.set_value('true')


class VarExtension(core.QTIElement, ExtendableExpressionMixin):

    """This element contains proprietary extensions to be applied to condition
    tests. This enables vendors to create their own conditional tests to be
    used on the participant responses::

            <!ELEMENT var_extension ANY>"""
    XMLNAME = "var_extension"
    XMLCONTENT = xml.XMLMixedContent


class PresentationMaterial(FlowMatContainer):

    """This is material that must be presented to set the context of the parent
    evaluation. This could be at the Section level to contain common question
    material that is relevant to all of the contained Sections/Items. All the
    contained material must be presented::

            <!ELEMENT presentation_material (qticomment? , flow_mat+)>

    Our interpretation is generous here, we also accept <material> by default
    from :py:class:`FlowMatContainer`.  This element is one of the newer
    definitions in QTI v1, after the introduction of <flow>.  It excludes
    <material> because it was assumed there would no legacy content.  Adoption
    of flow was poor and it was replaced with direct inclusion of the html
    model in version 2 (where content is either inline or block level and flow
    is a general term to describe both for contexts where either is
    allowed)."""
    XMLNAME = "presentation_material"
    XMLCONTENT = xml.ElementContent


class Reference(ContentMixin, QTICommentContainer):

    """The container for all of the materials that can be referenced by
    other structures e.g. feedback material, presentation material etc.
    The presentation of this material is under the control of the
    structure that is referencing the material. There is no implied
    relationship between any of the contained material components::

        <!ELEMENT reference (qticomment? , (material | mattext |
            matemtext | matimage | mataudio | matvideo | matapplet |
            matapplication | matbreak | mat_extension)+)>"""
    XMLNAME = "reference"
    XMLCONTENT = xml.ElementContent

    def content_child(self, child_class):
        return issubclass(
            child_class,
            (Material, MatText, MatEmText, MatImage, MatAudio, MatVideo,
             MatApplet, MatApplication, MatBreak, MatExtension))


class MaterialRef(core.QTIElement):

    """The <material_ref> element is used to contain a reference to the required
    full material block. This material will have had an identifier assigned to
    enable such a reference to be reconciled when the instance is parsed into
    the system::

            <!ELEMENT material_ref EMPTY>
            <!ATTLIST material_ref  linkrefid CDATA  #REQUIRED >"""
    XMLNAME = "material_ref"
    XMLATTR_linkrefid = 'linkRefID'
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.linkRefID = None

    def find_material(self):
        material = None
        doc = self.get_document()
        if doc:
            material = doc.find_material(self.linkRefID)
        if material is None:
            raise core.QTIError("Bad material_ref: %s" % str(self.linkRefID))
        return material

    def is_inline(self):
        return self.find_material().is_inline()

    def extract_text(self):
        return self.find_material().extract_text()

    def migrate_content_to_v2(self, parent, child_type, log):
        self.find_material().migrate_content_to_v2(parent, child_type, log)



#! /usr/bin/env python
"""Experimental module for Python 2.6 compatibility."""

import io
import logging
import sys
import zipfile

from wsgiref.simple_server import ServerHandler

from .py2 import is_text, builtins


py26 = sys.hexversion < 0x02070000
"""If you must know whether or not you are running under Python 2.6
then you can check using this flag,which is True in that case."""


def finish_content(self):
    # ugly patch for python 2.6; see the following for details:
    # http://stackoverflow.com/questions/3857029/does-wsgi-override-content-length-header
    if not self.headers_sent:
        if (self.environ.get('REQUEST_METHOD', '') != 'HEAD' or
                'Content-Length' not in self.headers):
            self.headers['Content-Length'] = 0
        self.send_headers()


def is_zipfile(filename):
    # patched to accept a file that is already open we use the
    # _EndRecData method directly, we're all grown-ups here
    try:
        if is_text(filename):
            fpin = open(filename, "rb")
            endrec = zipfile._EndRecData(fpin)
            fpin.close()
        else:
            endrec = zipfile._EndRecData(filename)
        if endrec:
            return True
    except IOError:
        pass
    return False


if py26:
    def get_method_function(methodlike):
        if isinstance(methodlike, classmethod):
            # second arg just needs to be any type
            return methodlike.__get__(None, type).im_func
        elif isinstance(methodlike, staticmethod):
            return methodlike.__get__(None, type)
        else:
            return methodlike
else:
    def get_method_function(methodlike):
        if isinstance(methodlike, (classmethod, staticmethod)):
            return methodlike.__func__
        else:
            return methodlike


if py26:
    logging.info("Adding missing constants to py26.io")
    io.SEEK_SET = 0
    io.SEEK_CUR = 1
    io.SEEK_END = 2

    class memoryview(object):       # noqa

        def __init__(self):
            raise TypeError("memoryview object not available in py26")

    logging.info("Patching wsgiref.simple_server.ServerHandler for HEAD bug")
    ServerHandler.finish_content = finish_content

    logging.info("Patching zipfile.is_zipfile for open files")
    zipfile.is_zipfile = is_zipfile
else:
    memoryview = builtins.memoryview


if py26:
    class RawIOBase(io.RawIOBase):

        def read(self, n=-1):
            """Read and return up to n bytes.

            Returns an empty bytes array on EOF, or None if the object is
            set not to block and has no data to read.

            Adapted from python2.6/io.py to deal with failure to return
            None in non-blocking case."""
            if n is None:
                n = -1
            if n < 0:
                return self.readall()
            b = bytearray(n.__index__())
            n = self.readinto(b)
            if n is None:
                return None
            else:
                del b[n:]
                return bytes(b)
else:
    RawIOBase = io.RawIOBase



#! /usr/bin/env python
"""Experimental module for Python 2 compatibility.

The purpose of this module is to enable Pyslet to be gradually converted
to Python3 while retaining support for Python 2.7 and 2.6.  This fills a
similar role to the six module but the idea is to minimise the number of
required fixes by making the Pyslet code as Python3 native as
possible."""

import io
import sys
import types

py2 = sys.hexversion < 0x03000000
"""Unfortunately, sometimes you just need to know if you are running
under Python 2, this flag provides a common way for version specific
code to check.  (There are multiple ways of checking, this flag just
makes it easier to find places in Pyslet where we care.)"""


_sys_codec = sys.getdefaultencoding()


if py2:
    suffix = ''

    def u8(arg):
        if isinstance(arg, types.UnicodeType):
            try:
                arg.encode('ascii')
            except UnicodeEncodeError:
                raise ValueError("u8: use binary literal for non-ASCII data")
            return arg
        try:
            return arg.decode('utf-8')
        except UnicodeDecodeError:
            raise ValueError("u8: invalid utf-8 string, did you mean ul?")

    def ul(arg):
        if isinstance(arg, types.UnicodeType):
            try:
                arg.encode('latin-1')
            except UnicodeEncodeError:
                raise ValueError("ul: cannot be used with non-latin data")
            return arg
        return arg.decode('latin-1')

    def is_string(arg):
        return isinstance(arg, types.StringTypes)

    is_text = is_string

    def force_text(arg):
        if isinstance(arg, str):
            return unicode(arg)
        elif isinstance(arg, unicode):
            return arg
        else:
            raise TypeError("Expected str or unicode: %s" % repr(arg))

    def is_ascii(arg):
        return isinstance(arg, str)

    def force_ascii(arg):
        if isinstance(arg, unicode):
            return arg.encode('ascii')
        elif isinstance(arg, str):
            return arg
        else:
            raise TypeError("Expected str or unicode: %s" % repr(arg))

    to_text = unicode

    def is_unicode(arg):
        return isinstance(arg, unicode)

    def character(arg):
        if isinstance(arg, str):
            if len(arg) == 1:
                return unichr(ord(arg[0]))
            else:
                raise ValueError('Expected single character')
        else:
            return unichr(arg)

    join_characters = unicode('').join

    uempty = unicode('')

    uspace = unicode(' ')

    def force_bytes(arg):
        if isinstance(arg, unicode):
            return arg.encode('ascii')
        return arg

    to_bytes = str

    def is_byte(arg):
        return isinstance(arg, bytes) and len(arg) == 1

    def byte(arg):
        if isinstance(arg, str):
            if len(arg) == 1:
                return arg
            else:
                raise ValueError('Expected single character')
        elif isinstance(arg, types.UnicodeType):
            if len(arg) == 1:
                arg = ord(arg)
                # fall through to int tests
            else:
                raise ValueError('Expected single character')
        elif isinstance(arg, bytearray):
            if len(arg) == 1:
                return chr(arg[0])
            else:
                raise ValueError('Expected single byte')
        if isinstance(arg, (int, long)):
            if arg >= 0 and arg <= 255:
                return chr(arg)
            else:
                raise ValueError("Value out of range 0..255")
        else:
            raise TypeError('Expectected character or int')

    byte_value = ord

    join_bytes = b''.join

    def byte_to_bstr(arg):
        return arg

    buffer2 = types.BufferType

    long2 = long

    range3 = xrange

    def dict_keys(d):
        return d.iterkeys()

    def dict_values(d):
        return d.itervalues()

    def dict_items(d):
        return d.iteritems()

    import __builtin__ as builtins

    input3 = raw_input

    from urllib import (            # noqa : unused import
        urlencode,
        urlopen,
        quote as urlquote
        )
    from urlparse import parse_qs   # noqa : unused import

else:
    suffix = '3'

    def u8(arg):
        if isinstance(arg, bytes):
            return arg.decode('utf-8')
        elif isinstance(arg, str):
            # only works for ascii
            try:
                arg.encode('ascii')
            except UnicodeEncodeError:
                raise ValueError("u8: use binary literal for non-ASCII data")
            return arg
        else:
            raise TypeError

    def ul(arg):
        if isinstance(arg, bytes):
            return arg.decode('latin-1')
        elif isinstance(arg, str):
            try:
                arg.encode('latin-1')
            except UnicodeEncodeError:
                raise ValueError("ul: cannot be used with non-latin data")
            return arg
        else:
            raise TypeError

    def is_string(arg):
        return isinstance(arg, (str, bytes))

    def is_text(arg):
        return isinstance(arg, str)

    def force_text(arg):
        if not isinstance(arg, str):
            raise TypeError("Expected str: %s" % repr(arg))
        return arg

    def is_ascii(arg):
        if isinstance(arg, str):
            arg.encode('ascii')
            return True
        else:
            return False

    def force_ascii(arg):
        if isinstance(arg, bytes):
            return arg.decode('ascii')
        elif isinstance(arg, str):
            return arg
        else:
            raise TypeError("Expected str: %s" % repr(arg))

    def to_text(arg):
        if isinstance(arg, str):
            return arg
        elif isinstance(arg, bytes):
            return arg.decode('ascii')
        else:
            return str(arg)

    def is_unicode(arg):
        return isinstance(arg, str)

    character = chr

    join_characters = ''.join

    uempty = ''

    uspace = ' '

    def force_bytes(arg):
        if isinstance(arg, str):
            return arg.encode('ascii')
        return arg

    def to_bytes(arg):
        if hasattr(arg, '__bytes__'):
            return arg.__bytes__()
        else:
            return str(arg).encode('ascii')

    def is_byte(arg):
        return isinstance(arg, int) and 0 <= arg <= 255

    def byte(arg):
        if isinstance(arg, str):
            if len(arg) == 1:
                arg = ord(arg)
            else:
                raise ValueError('Expected single character')
        elif isinstance(arg, (bytes, bytearray)):
            if len(arg) == 1:
                arg = arg[0]
            else:
                raise ValueError('Expected single byte')
        if isinstance(arg, int):
            if arg >= 0 and arg <= 255:
                return arg
            else:
                raise ValueError("Value out of range 0..255")
        else:
            raise TypeError('Expectected character or int')

    byte_value = int

    join_bytes = bytes

    def byte_to_bstr(arg):
        return bytes([arg])

    buffer2 = bytes

    long2 = int

    range3 = range

    def dict_keys(d):
        return d.keys()

    def dict_values(d):
        return d.values()

    def dict_items(d):
        return d.items()

    import builtins     # noqa : unused import

    input3 = input

    from urllib.request import urlopen      # noqa : unused import
    from urllib.parse import (              # noqa : unused import
        parse_qs,
        quote as urlquote,
        urlencode
        )


class UnicodeMixin(object):

    """Mixin class to handle string formatting

    For classes that need to define a __unicode__ method of their own
    this class is used to ensure that the correct behaviour exists
    in Python versions 2 and 3.

    The mixin class implements __str__ based on your existing (required)
    __unicode__ or (optional) __bytes__ implementation.  In python 2,
    the output of __unicode__ is encoded using the default system
    encoding if no __bytes__ implementation is provided.  This may well
    generate errors but that seems more appropriate as it will catch
    cases where the *str* function has been used instead of
    :py:func:`to_text`."""

    if py2:
        def __str__(self):      # noqa
            if hasattr(self, '__bytes__'):
                return self.__bytes__()
            else:
                return self.__unicode__().encode(_sys_codec)
    else:
        def __str__(self):      # noqa
            return self.__unicode__()


class SortableMixin(object):

    """Mixin class for handling comparisons

    Utility class for helping provide comparisons that are compatible
    with Python 2 and Python 3.  Classes must define a method
    :meth:`sortkey` which returns a sortable key value representing the
    instance.

    Derived classes may optionally override the classmethod :meth:`otherkey`
    to provide an ordering against other object types.

    This mixin then adds implementations for all of the comparison
    methods: __eq__, __ne__, __lt__, __le__, __gt__, __ge__."""

    def sortkey(self):
        """Returns a value to use as a key for sorting.

        By default returns NotImplemented.  This value causes the
        comparison functions to also return NotImplemented."""
        return NotImplemented

    def otherkey(self, other):
        """Returns a value to use as a key for sorting

        The difference between this method and :meth:`sortkey` is that
        this method takes an arbitrary object and either returns the key
        to use when comparing with this instance or NotImplemented if
        the sorting is not supported.

        You don't have to override this implementation, by default it
        returns other.sortkey() if *other* is an instance of the same
        class as *self*, otherwise it returns NotImplemented."""
        if isinstance(other, self.__class__):
            return other.sortkey()
        else:
            return NotImplemented

    def __eq__(self, other):
        a = self.sortkey()
        b = self.otherkey(other)
        if NotImplemented in (a, b):
            return NotImplemented
        else:
            return a == b

    def __ne__(self, other):
        a = self.sortkey()
        b = self.otherkey(other)
        if NotImplemented in (a, b):
            return NotImplemented
        else:
            return a != b

    def __lt__(self, other):
        a = self.sortkey()
        b = self.otherkey(other)
        if NotImplemented in (a, b):
            if py2:
                raise TypeError("unorderable types: %s < %s" %
                                (repr(self), repr(other)))
            return NotImplemented
        else:
            return a < b

    def __le__(self, other):
        a = self.sortkey()
        b = self.otherkey(other)
        if NotImplemented in (a, b):
            if py2:
                raise TypeError("unorderable types: %s <= %s" %
                                (repr(self), repr(other)))
            return NotImplemented
        else:
            return a <= b

    def __gt__(self, other):
        a = self.sortkey()
        b = self.otherkey(other)
        if NotImplemented in (a, b):
            if py2:
                raise TypeError("unorderable types: %s > %s" %
                                (repr(self), repr(other)))
            return NotImplemented
        else:
            return a > b

    def __ge__(self, other):
        a = self.sortkey()
        b = self.otherkey(other)
        if NotImplemented in (a, b):
            if py2:
                raise TypeError("unorderable types: %s >= %s" %
                                (repr(self), repr(other)))
            return NotImplemented
        else:
            return a >= b


class CmpMixin(object):

    """Mixin class for handling comparisons

    For compatibility with Python 2's __cmp__ method this class defines
    an implementation of __eq__, __lt__, __le__, __gt__, __ge__ that are
    redirected to __cmp__.  These are the minimum methods required for
    Python's rich comparisons.

    In Python 2 it also provides an implementation of __ne__ that simply
    inverts the result of __eq__.  (This is not required in Python 3.)"""

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __le__(self, other):
        return self.__cmp__(other) <= 0

    def __gt__(self, other):
        return self.__cmp__(other) > 0

    def __ge__(self, other):
        return self.__cmp__(other) >= 0

    if py2:
        def __ne__(self, other):    # noqa
            return not self.__eq__(other)


class BoolMixin(object):

    """Mixin class for handling legacy __nonzero__

    For compatibility with Python 2 this class defines __nonzero__
    returning the value of the method __bool__."""

    def __nonzero__(self):
        return self.__bool__()


def output(txt):

    """Simple function for writing to stdout

    Not as sophisticated as Python 3's print function but designed to be
    more of a companion to the built in input."""
    if isinstance(sys.stdout, io.TextIOBase):
        sys.stdout.write(txt)
    else:
        sys.stdout.write(txt.encode('utf-8'))



#! /usr/bin/env python

import itertools

from . import common
from . import core
from ..xml import structures as xml


class Section(core.ObjectMixin, core.SectionMixin, common.QTICommentContainer):

    """The Section data structure is used to define arbitrarily complex
    hierarchical section and item data structures. It may contain meta-data,
    objectives, rubric control switches, assessment-level processing, feedback
    and selection and sequencing information for sections and items::

    <!ELEMENT section (qticomment? ,
            duration? ,
            qtimetadata* ,
            objectives* ,
            sectioncontrol* ,
            sectionprecondition* ,
            sectionpostcondition* ,
            rubric* ,
            presentation_material? ,
            outcomes_processing* ,
            sectionproc_extension? ,
            sectionfeedback* ,
            selection_ordering? ,
            reference? ,
            (itemref | item | sectionref | section)*
            )>
    <!ATTLIST section  ident CDATA  #REQUIRED
            title CDATA  #IMPLIED
            xml:lang CDATA  #IMPLIED >"""
    XMLNAME = "section"
    XMLATTR_ident = 'ident'
    XMLATTR_title = 'title'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        self.ident = None
        self.title = None
        self.Duration = None
        self.QTIMetadata = []
        self.Objectives = []
        self.SectionControl = []
        self.SectionPrecondition = []
        self.SectionPostcondition = []
        self.Rubric = []
        self.PresentationMaterial = None
        self.QTIOutcomesProcessing = []
        self.SectionProcExtension = None
        self.SectionFeedback = []
        self.QTISelectionOrdering = None
        self.QTIReference = None
        self.SectionItemMixin = []

    def get_children(self):
        for child in itertools.chain(
                core.QTIComment.get_children(self),
                self.QTIMetadata,
                self.Objectives,
                self.SectionControl,
                self.SectionPrecondition,
                self.SectionPostcondition,
                self.Rubric):
            yield child
        if self.PresentationMaterial:
            yield self.PresentationMaterial
        for child in self.QTIOutcomesProcessing:
            yield child
        if self.SectionProcExtension:
            yield self.SectionProcExtension
        for child in self.SectionFeedback:
            yield child
        if self.QTISelectionOrdering:
            yield self.QTISelectionOrdering
        if self.QTIReference:
            yield self.QTIReference
        for child in self.SectionItemMixin:
            yield child

    def migrate_to_v2(self, output):
        """Converts this section to QTI v2

        For details, see
        :py:class:`pyslet.qtiv1.QuesTestInterop.migrate_to_v2`."""
        for obj in self.SectionItemMixin:
            obj.migrate_to_v2(output)


class SectionPrecondition(core.QTIElement):

    """The preconditions that control whether or not the Section is activated.
    This is for further study in V2.0::

    <!ELEMENT sectionprecondition (#PCDATA)>"""
    XMLNAME = 'sectionprecondition'
    XMLCONTENT = xml.XMLMixedContent


class SectionPostcondition(core.QTIElement):

    """The postconditions that control whether or not the next Section may be
    activated. This is for further study in V2.0::

    <!ELEMENT sectionpostcondition (#PCDATA)>"""
    XMLNAME = 'sectionpostcondition'
    XMLCONTENT = xml.XMLMixedContent


class SectionControl(common.QTICommentContainer):

    """The control switches that are used to enable or disable the display of
    hints, solutions and feedback within the Section::

    <!ELEMENT sectioncontrol (qticomment?)>
    <!ATTLIST sectioncontrol  feedbackswitch  (Yes | No )  'Yes'
            hintswitch  (Yes | No )  'Yes'
            solutionswitch  (Yes | No )  'Yes'
            view	(All | Administrator | AdminAuthority | Assessor | Author |
                            Candidate | InvigilatorProctor | Psychometrician |
                            Scorer | Tutor ) 'All' >"""
    XMLNAME = 'sectioncontrol'
    XMLATTR_feedbackswitch = (
        'feedbackSwitch', core.ParseYesNo, core.FormatYesNo)
    XMLATTR_hintswitch = ('hintSwitch', core.ParseYesNo, core.FormatYesNo)
    XMLATTR_solutionswitch = (
        'solutionSwitch', core.ParseYesNo, core.FormatYesNo)
    XMLATTR_view = ('view', core.View.from_str_lower, core.View.to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        self.feedbackSwitch = True
        self.hintSwitch = True
        self.solutionSwitch = True
        self.view = core.View.DEFAULT


class SectionProcExtension(core.QTIElement):

    """This is used to contain proprietary alternative Section-level processing
    functionality::

    <!ELEMENT sectionproc_extension ANY>"""
    XMLNAME = "sectionproc_extension"
    XMLCONTENT = xml.XMLMixedContent


class SectionFeedback(common.ContentMixin, common.QTICommentContainer):

    """The container for the Section-level feedback that is to be presented as a
    result of Section-level processing of the user's responses::

    <!ELEMENT sectionfeedback (qticomment? , (material+ | flow_mat+))>
    <!ATTLIST sectionfeedback
            view	(All | Administrator | AdminAuthority | Assessor | Author |
                            Candidate | InvigilatorProctor | Psychometrician |
                            Scorer | Tutor ) 'All'
            ident CDATA  #REQUIRED
            title CDATA  #IMPLIED >"""
    XMLNAME = 'sectionfeedback'
    XMLATTR_view = ('view', core.View.from_str_lower, core.View.to_str)
    XMLATTR_ident = 'ident'
    XMLATTR_title = 'title'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        common.ContentMixin.__init__(self)
        self.view = core.View.DEFAULT
        self.ident = None
        self.title = None

    def get_children(self):
        return itertools.chain(
            common.QTICommentContainer.get_children(self),
            common.ContentMixin.get_content_children(self))

    def content_child(self, child_class):
        return child_class in (common.Material, common.FlowMat)


class SectionRef(core.SectionMixin, core.QTIElement):

    """Represents the sectionref element::

    <!ELEMENT sectionref (#PCDATA)>
    <!ATTLIST sectionref  linkrefid CDATA  #REQUIRED >"""
    XMLNAME = 'sectionref'
    XMLATTR_linkrefid = 'linkrefid'
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.linkrefid = None

    def migrate_to_v2(self, output):
        """Converts this sectionref to QTI v2

        Currently does nothing."""
        pass


class ItemRef(core.QTIElement, core.SectionItemMixin):

    """This element is used to 'pull' an Item into the scope. This is used to
    refer to a Item that has been defined elsewhere but which is to be
    logically related to this group of Items::

    <!ELEMENT itemref (#PCDATA)>
    <!ATTLIST itemref  linkrefid CDATA  #REQUIRED >"""
    XMLNAME = 'itemref'
    XMLATTR_linkrefid = 'linkrefid'
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.linkrefid = None

    def migrate_to_v2(self, output):
        """Converts this itemref to QTI v2

        Currently does nothing."""
        pass



#! /usr/bin/env python

import warnings

from pyslet.http.client import *        # noqa


warnings.warn("rfc2616 module is deprecated, use http package instead",
              DeprecationWarning,
              stacklevel=2)
